from datetime import datetime
from datetime import timedelta

import factory
from factory import fuzzy
from pony import orm
from pyoadr_ven import database


class EventFactory(factory.Factory):
    class Meta:
        model = database.Event

    request_id = factory.Sequence(lambda n: str(n))
    event_id = factory.Sequence(lambda n: str(n))

    official_start = fuzzy.FuzzyNaiveDateTime(
        datetime.utcnow() + timedelta(days=+1), datetime.utcnow() + timedelta(days=+14)
    )
    duration = timedelta(minutes=30)
    start_after = timedelta(seconds=0)
    start_offset = timedelta(seconds=0)
    start_time = factory.LazyAttribute(lambda e: e.official_start + e.start_offset)
    end_time = factory.LazyAttribute(lambda e: e.official_start + e.duration)

    @classmethod
    def _create(cls, model_class, *args, **kwargs):
        obj = model_class(*args, **kwargs)
        orm.commit()
        return obj


class ReportFactory(factory.Factory):
    class Meta:
        model = database.Report

    start_time = datetime.utcnow()
    end_time = datetime.utcnow() + timedelta(hours=12)
    duration = timedelta(hours=12)
    name = "test_report"
    interval_secs = 1
    telemetry_parameters = ""
    request_id = factory.Sequence(lambda n: str(n))
    specifier_id = factory.Sequence(lambda n: str(n))

    @classmethod
    def _create(cls, model_class, *args, **kwargs):
        obj = model_class(*args, **kwargs)
        orm.commit()
        return obj



import logging
from unittest.mock import MagicMock

import pytest
import responses

_LOGGER = logging.getLogger(__name__)

VTN_ADDRESS = "https://openadr-staging"
ENDPOINT_BASE = "/OpenADR2/Simple/2.0b/"
ENDPOINT = VTN_ADDRESS + ENDPOINT_BASE
EIEVENT = ENDPOINT + "EiEvent"
EIREPORT = ENDPOINT + "EiReport"
EIREGISTERPARTY = ENDPOINT + "EiRegisterParty"
POLL = ENDPOINT + "OadrPoll"

with open("tests/xml/vtn_distribute_event.xml", "r") as f:
    distribute_event_response_body = f.read()

with open("tests/xml/vtn_empty_poll_response.xml", "r") as f:
    empty_poll_response_body = f.read()

with open("tests/xml/ven_created_event.xml", "r") as f:
    ven_created_event = f.read()

with open("tests/xml/test_vtn_cancel_event.xml", "r") as f:
    canceled_event_response_body = f.read()

pytestmark = pytest.mark.pony(db_session=False)


class TestEmptyPoll:
    @responses.activate
    def test_returns_no_events(self, mocked_agent):
        responses.add(
            responses.POST, POLL, body=empty_poll_response_body, content_type="text/xml"
        )
        mocked_agent.send_oadr_poll()
        _LOGGER.warning(responses.calls[:])
        assert responses.calls[0].request.url == POLL
        assert mocked_agent._active_or_pending_events == []


class TestPollWithEvent:
    @responses.activate
    def test_adds_events_to_db(self, mocked_agent):
        responses.add(
            responses.POST,
            POLL,
            content_type="text/xml",
            body=distribute_event_response_body,
        )
        responses.add(
            responses.POST, POLL, content_type="text/xml", body=empty_poll_response_body
        )
        responses.add(responses.POST, EIEVENT, content_type="text/xml")

        mocked_agent.send_oadr_poll()
        assert responses.calls[0].request.url == POLL
        assert responses.calls[1].request.url == EIEVENT
        assert responses.calls[2].request.url == EIEVENT
        assert responses.calls[3].request.url == POLL

        assert len(mocked_agent._active_or_pending_events) == 1
        assert len(mocked_agent._far_events) == 1
        assert len(mocked_agent._near_events) == 0
        assert len(mocked_agent._active_events) == 0
        assert len(mocked_agent._unresponded_events) == 0


class TestRunMainProcessesWithNoEvents:
    @responses.activate
    def test_calls_send_oadr_poll(self, mocked_agent):
        mocked_agent.send_oadr_poll = MagicMock()
        mocked_agent.tick()
        mocked_agent.send_oadr_poll.assert_called()

    @responses.activate
    def test_calls_process_event(self, mocked_agent):
        mocked_agent.process_event = MagicMock()
        mocked_agent.tick()
        mocked_agent.process_event.assert_not_called()


class TestRunMainProcessesWithEvents:
    @responses.activate
    def test_calls_send_oadr_poll(self, mocked_agent):
        responses.add(
            responses.POST,
            POLL,
            content_type="text/xml",
            body=distribute_event_response_body,
        )
        responses.add(
            responses.POST, POLL, content_type="text/xml", body=empty_poll_response_body
        )
        responses.add(responses.POST, EIEVENT, content_type="text/xml")
        mocked_agent.send_oadr_poll = MagicMock()
        mocked_agent.tick()
        mocked_agent.send_oadr_poll.assert_called()

    @responses.activate
    def test_calls_process_event(self, mocked_agent):
        responses.add(
            responses.POST,
            POLL,
            content_type="text/xml",
            body=distribute_event_response_body,
        )
        responses.add(
            responses.POST, POLL, content_type="text/xml", body=empty_poll_response_body
        )
        responses.add(responses.POST, EIEVENT, content_type="text/xml")
        mocked_agent.process_event = MagicMock()
        mocked_agent.tick()
        mocked_agent.process_event.assert_called()


class TestEventCancelation:
    @responses.activate
    def test_event_cancelation(self, mocked_agent):
        responses.add(
            responses.POST,
            POLL,
            content_type="text/xml",
            body=distribute_event_response_body,
        )
        responses.add(
            responses.POST, POLL, content_type="text/xml", body=empty_poll_response_body
        )
        responses.add(responses.POST, EIEVENT, content_type="text/xml")

        mocked_agent.send_oadr_poll()
        assert responses.calls[0].request.url == POLL
        assert responses.calls[1].request.url == EIEVENT
        assert responses.calls[2].request.url == EIEVENT
        assert responses.calls[3].request.url == POLL

        assert len(mocked_agent._active_or_pending_events) == 1
        assert len(mocked_agent._far_events) == 1
        assert len(mocked_agent._near_events) == 0
        assert len(mocked_agent._active_events) == 0
        assert len(mocked_agent._unresponded_events) == 0

        responses.add(
            responses.POST,
            POLL,
            content_type="text/xml",
            body=canceled_event_response_body,
        )
        mocked_agent.send_oadr_poll()
        assert responses.calls[0].request.url == POLL
        for event in mocked_agent._active_or_pending_events:
            mocked_agent.process_event(event)

        assert len(mocked_agent._active_or_pending_events) == 0



import os

import pytest

from pyoadr_ven import database
from pyoadr_ven import OpenADRVenAgent


LOCAL_VTN_ADDRESS = "http://localhost:3000"
LOCAL_VEN_ID = "1"
LOCAL_VTN_ID = "ccoopvtn01"

MOCK_VTN_ADDRESS = "https://openadr-staging"
MOCK_VEN_ID = "ven01"
MOCK_VTN_ID = "vtn01"
test_dir = os.path.dirname(os.path.abspath(__file__))
CERT_PATH = test_dir + "/client.pem"
CA_PATH = test_dir + "/ca.crt"

database.setup_db()

evese_telemetry_short_status_parameters = {
    "state": {
        "r_id": "evse state",
        "units": "NA",
        "min_frequency": 30,
        "max_frequency": 30,
        "report_type": "",
        "reading_type": "",
        "method_name": "state",
    },
    "amp": {
        "r_id": "evse charging current",
        "units": "A",
        "min_frequency": 30,
        "max_frequency": 30,
        "report_type": "",
        "reading_type": "",
        "method_name": "amp",
    },
    "wh": {
        "r_id": "evse energy used in session",
        "units": "Wh",
        "min_frequency": 30,
        "max_frequency": 30,
        "report_type": "",
        "reading_type": "",
        "method_name": "wh",
    },
}

evse_telemetry_short_status_report_parameters = {
    "ccoop_telemetry_evse_status": {
        "report_name_metadata": "ccoop_telemetry_evse_status",
        "report_interval_secs_default": 30,
        "telemetry_parameters": evese_telemetry_short_status_parameters,
    }
}


@pytest.fixture
def mocked_agent():

    agent = OpenADRVenAgent(
        ven_id=MOCK_VEN_ID,
        vtn_id=MOCK_VTN_ID,
        vtn_address=MOCK_VTN_ADDRESS,
        security_level="standard",
        poll_interval_secs=15,
        log_xml=True,
        opt_timeout_secs=30,
        opt_default_decision="optIn",
        report_parameters=evse_telemetry_short_status_report_parameters,
        client_pem_bundle=CERT_PATH,
        vtn_ca_cert=CA_PATH,
    )
    return agent


@pytest.fixture
def local_agent():
    agent = OpenADRVenAgent(
        ven_id=LOCAL_VEN_ID,
        vtn_id=LOCAL_VTN_ID,
        vtn_address=LOCAL_VTN_ADDRESS,
        security_level="standard",
        poll_interval_secs=15,
        log_xml=True,
        opt_timeout_secs=30,
        opt_default_decision="optIn",
        report_parameters=evse_telemetry_short_status_report_parameters,
        client_pem_bundle=CERT_PATH,
        vtn_ca_cert=CA_PATH,
    )
    return agent



import json
import logging
from datetime import datetime

import pytest

from pyoadr_ven import database


_LOGGER = logging.getLogger(__name__)
pytestmark = pytest.mark.pony


class TestTelemetry:
    def test_add_telemetry_json(self, mocked_agent):
        values_dict = {"state": "on"}
        mocked_agent.add_telemetry_json(
            "ccoop_telemetry_evse_status",
            start_time=datetime.utcnow(),
            values=values_dict,
        )
        assert database.TelemetryValues[1].values == json.dumps(values_dict)
        assert (
            database.TelemetryValues[1].report_specifier_id
            == "ccoop_telemetry_evse_status"
        )

    # def test_get_new_telemetry_for_report(self, mocked_agent):
    #     mocked_agent.get_new_telemetry_for_report()



import logging

from pyoadr_ven import OpenADRVenAgent

logging.basicConfig(level=logging.DEBUG)
_LOGGER = logging.getLogger(__name__)
evese_telemetry_short_status_parameters = {
    "state": {
        "r_id": "evse state",
        "units": "NA",
        "min_frequency": 30,
        "max_frequency": 30,
        "report_type": "",
        "reading_type": "",
        "method_name": "state",
    },
    "amp": {
        "r_id": "evse charging current",
        "units": "A",
        "min_frequency": 30,
        "max_frequency": 30,
        "report_type": "",
        "reading_type": "",
        "method_name": "amp",
    },
    "wh": {
        "r_id": "evse energy used in session",
        "units": "Wh",
        "min_frequency": 30,
        "max_frequency": 30,
        "report_type": "",
        "reading_type": "",
        "method_name": "wh",
    },
}

evse_telemetry_short_status_report_parameters = {
    "ccoop_telemetry_evse_status": {
        "report_name": "ccoop_telemetry_evse_status",
        "report_interval_secs_default": 30,
        "telemetry_parameters": evese_telemetry_short_status_parameters,
    }
}

agent = OpenADRVenAgent(
    ven_id="1",
    vtn_id="ccoopvtn01",
    vtn_address="http://localhost:3000",
    client_pem_bundle="/home/peter/carboncoop/pyoadr-ven/tests/client.pem",
    vtn_ca_cert="/home/peter/carboncoop/pyoadr-ven/tests/ca.crt",
    report_parameters=evse_telemetry_short_status_report_parameters,
    log_xml=True,
    db_filepath="/home/peter/carboncoop/pyoadr-ven/db.sqlite",
)

agent.send_oadr_register_report()



HTTP_STATUS_CODES = {
    200: "200 OK",
    201: "201 Created",
    204: "204 No Content",
    500: "500 Internal Error",
}

# OADR Response Codes
OADR_VALID_RESPONSE = "200"
OADR_MOD_NUMBER_OUT_OF_ORDER = "450"
OADR_NOT_RECOGNISED = "453"
OADR_INVALID_DATA = "454"
OADR_BAD_DATA = "459"
OADR_BAD_SIGNAL = "460"



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2017, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}
###################################################
#
# Changes for HomeAssistant are marked with **HA**
#
###################################################
"""VOLTTRON platform™ agent helper classes/functions."""
import logging
from datetime import datetime

import pytz

_LOGGER = logging.getLogger(__name__)


def format_timestamp(time_stamp):
    """Create a consistent datetime string representation based on
    ISO 8601 format.

    YYYY-MM-DDTHH:MM:SS.mmmmmm for unaware datetime objects.
    YYYY-MM-DDTHH:MM:SS.mmmmmm+HH:MM for aware datetime objects

    :param time_stamp: value to convert
    :type time_stamp: datetime
    :returns: datetime in string format
    :rtype: str
    """

    time_str = time_stamp.strftime("%Y-%m-%dT%H:%M:%S.%f")

    if time_stamp.tzinfo is not None:
        sign = "+"
        td = time_stamp.tzinfo.utcoffset(time_stamp)
        if td.days < 0:
            sign = "-"
            td = -td

        seconds = td.seconds
        minutes, seconds = divmod(seconds, 60)
        hours, minutes = divmod(minutes, 60)
        time_str += "{sign}{HH:02}:{MM:02}".format(sign=sign, HH=hours, MM=minutes)

    return time_str


def get_aware_utc_now():
    """Create a timezone aware UTC datetime object from the system time.

    :returns: an aware UTC datetime object
    :rtype: datetime
    """
    utcnow = datetime.utcnow()
    utcnow = pytz.UTC.localize(utcnow)
    return utcnow



import logging
import time
from datetime import datetime
from datetime import timedelta
from unittest.mock import MagicMock

import pytest
from time_machine import travel
from pony import orm
from voluptuous import Invalid

from pyoadr_ven import database
from pyoadr_ven import OpenADRVenAgent
from pyoadr_ven.enums import ReportStatus
from tests.conftest import CA_PATH
from tests.conftest import CERT_PATH
from tests.factories import ReportFactory

_LOGGER = logging.getLogger(__name__)

pytestmark = pytest.mark.pony

TELEMETRY_PARAMS = {
    "state": {
        "r_id": "evse state",
        "units": "NA",
        "min_frequency": 30,
        "max_frequency": 30,
        "report_type": "",
        "reading_type": "",
        "method_name": "state",
    },
    "amp": {
        "r_id": "evse charging current",
        "units": "A",
        "min_frequency": 30,
        "max_frequency": 30,
        "report_type": "",
        "reading_type": "",
        "method_name": "amp",
    },
    "wh": {
        "r_id": "evse energy used in session",
        "units": "Wh",
        "min_frequency": 30,
        "max_frequency": 30,
        "report_type": "",
        "reading_type": "",
        "method_name": "wh",
    },
}

REPORT_PARAMS = {
    "ccoop_telemetry_evse_status": {
        "report_name_metadata": "ccoop_telemetry_evse_status",
        "report_interval_secs_default": 30,
        "telemetry_parameters": TELEMETRY_PARAMS,
    }
}

MOCK_VTN_ADDRESS = "https://openadr-staging"
MOCK_VEN_ID = "ven01"
MOCK_VTN_ID = "vtn01"
LOCAL_VTN_ADDRESS = "http://localhost:3000"
LOCAL_VEN_ID = "1"
LOCAL_VTN_ID = "ccoopvtn01"


class TestReportParameters:
    def test_invalid_report_parameters_raise_exception(self):
        invalid_report_params = REPORT_PARAMS = {
            "ccoop_telemetry_evse_status": {
                "report_name_metadata": "ccoop_telemetry_evse_status",
                "report_interal_secs_default": 30,
                "telemetry_parameters": TELEMETRY_PARAMS,
            }
        }
        with pytest.raises(Invalid):
            agent = OpenADRVenAgent(
                ven_id=LOCAL_VEN_ID,
                vtn_id=LOCAL_VTN_ID,
                vtn_address=LOCAL_VTN_ADDRESS,
                security_level="standard",
                poll_interval_secs=5,
                log_xml=True,
                opt_timeout_secs=30,
                opt_default_decision="optIn",
                report_parameters=REPORT_PARAMS,
                client_pem_bundle=CERT_PATH,
                vtn_ca_cert=CA_PATH,
            )

    def test_blank_report_parameters_valid(self):

        agent = OpenADRVenAgent(
            ven_id=LOCAL_VEN_ID,
            vtn_id=LOCAL_VTN_ID,
            vtn_address=LOCAL_VTN_ADDRESS,
            security_level="standard",
            poll_interval_secs=5,
            log_xml=True,
            opt_timeout_secs=30,
            opt_default_decision="optIn",
            report_parameters={},
            client_pem_bundle=CERT_PATH,
            vtn_ca_cert=CA_PATH,
        )


class TestReportCreationRegistration:
    @pytest.mark.local
    def test_inactive_report_created_on_register_report(self, mocked_agent):

        mocked_agent.send_oadr_register_report()

        assert orm.select(r for r in database.Report).count() == 1

        test_report = orm.select(r for r in database.Report).first()
        assert test_report
        assert test_report.specifier_id == "ccoop_telemetry_evse_status"
        assert test_report.status == ReportStatus.INACTIVE.value

    @pytest.mark.local
    def test_handle_oadr_registered_report_works(self, caplog):
        """
        To run this test you need to ensure the following:
            #. Start your VTN server
                * note the url it comes up on - if it isn't localhost:3000,
                    change the LOCAL_VTN_ADDRESS variable

            #. Go to the Admin page of the VTN and create a certificate authority called ``default``
                * save the CA certificate in the tests folder as 'ca.crt'
            #. Still in the VTN Admin page, create a certificate
                * save the certificate and private key in the tests folder as 'client.pem'

            .. code-block:: bash

                -----BEGIN PRIVATE KEY-----
                MIIEvQIBADANBgkqhkiG9w0BAQEFAASCBKcwggSjAgEAAoIBAQCqFiHgHccwrsrs
                ...
                lTwad+cbPVyQMzCsxEl7e7A=
                -----END PRIVATE KEY-----
                -----BEGIN CERTIFICATE-----
                MIIESzCCAzOgAwIBAgIRAOY4YtDbjUM4gek4QkBib6cwDQYJKoZIhvcNAQELBQAw
                ...
                -----END CERTIFICATE-----


            #. Now, using the main VTN interface (not the Django admin), create a VEN
                on the VTN - you will need to create a Customer and Site to do this.
                * set the LOCAL_VEN_ID variable to the ven_id you used
                * Note the LOCAL_VTN_ID used (see base.py in the VTN repo)

            #. Using the Django admin, create a DRProgram on the VTN
                * Add your VEN/site to this program

        Each time you run the test, it is better to have a clean VTN database.
        Open a django shell and remove all of the Reports and Telemetry objects.

        """
        caplog.set_level(logging.INFO)
        agent = OpenADRVenAgent(
            ven_id=LOCAL_VEN_ID,
            vtn_id=LOCAL_VTN_ID,
            vtn_address=LOCAL_VTN_ADDRESS,
            security_level="standard",
            poll_interval_secs=5,
            log_xml=True,
            opt_timeout_secs=30,
            opt_default_decision="optIn",
            report_parameters=REPORT_PARAMS,
            client_pem_bundle=CERT_PATH,
            vtn_ca_cert=CA_PATH,
        )
        agent.send_oadr_register_report()
        agent.tick()
        # agent.reports_registered = True
        assert orm.select(r for r in database.Report).count() == 1
        test_report = orm.select(r for r in database.Report).first()
        agent.process_report(test_report.id)
        assert test_report.status == ReportStatus.ACTIVE.value
        agent.add_telemetry_json(
            report_specifier_id=test_report.specifier_id,
            values={"state": 1, "amp": 40, "wh": 50},
        )
        tel = agent.get_new_telemetry_for_report(
            report_specifier_id=test_report.specifier_id
        )
        assert len(tel) == 1
        # while True:
        #     time.sleep(6)
        #     agent.tick()


class TestReportStatusHelpers:
    def test_inactive_report_in_active_or_pending(self, mocked_agent):
        report = ReportFactory(status=ReportStatus.INACTIVE.value)
        assert report.status == ReportStatus.INACTIVE.value
        assert report.is_active_or_pending
        assert report in mocked_agent.active_or_pending_reports

    def test_active_report_in_active_or_pending(self, mocked_agent):
        report = ReportFactory(status=ReportStatus.ACTIVE.value)
        assert report.status == ReportStatus.ACTIVE.value
        assert report.is_active_or_pending
        assert report in mocked_agent.active_or_pending_reports

    def test_completed_report_not_in_active_or_pending(self, mocked_agent):
        report = ReportFactory(status=ReportStatus.COMPLETED.value)
        assert report.status == ReportStatus.COMPLETED.value
        assert report.is_active_or_pending is False
        assert report not in mocked_agent.active_or_pending_reports

    def test_canceled_report_not_in_active_or_pending(self, mocked_agent):
        report = ReportFactory(status=ReportStatus.CANCELED.value)
        assert report.status == ReportStatus.CANCELED.value
        assert report.is_active_or_pending is False
        assert report not in mocked_agent.active_or_pending_reports


class TestWithTimeMovingOn:
    @pytest.mark.local
    def test_active_report_becomes_complete(self, mocked_agent):
        initial_datetime = datetime(year=2019, month=1, day=1, hour=10, minute=0)
        with travel(initial_datetime) as traveller:
            report = ReportFactory(
                start_time=initial_datetime,
                end_time=datetime.utcnow() + timedelta(hours=1),
                created=initial_datetime,
                status=ReportStatus.ACTIVE.value,
            )
            mocked_agent.process_report(report.id)
            assert report.status == ReportStatus.ACTIVE.value
            traveller.tick(delta=timedelta(hours=2))
            mocked_agent.process_report(report.id)
            assert report.status == ReportStatus.COMPLETED.value



import logging
from datetime import datetime
from datetime import timedelta

import pytest
from time_machine import travel

from .factories import EventFactory
from pyoadr_ven import database
from pyoadr_ven import exceptions
from pyoadr_ven.enums import EventStatus
from pyoadr_ven.enums import OptType
from pyoadr_ven.enums import ReportStatus

_LOGGER = logging.getLogger(__name__)
pytestmark = pytest.mark.pony


class TestSetEventStatusTransition:
    def test_from_unresponded_to_completed(self, mocked_agent):
        event = EventFactory()

        assert event.status == EventStatus.UNRESPONDED.value
        assert mocked_agent._unresponded_events[0] == event
        assert mocked_agent._active_or_pending_events[0] == event
        assert database.Event.get(status=EventStatus.UNRESPONDED.value) == event
        mocked_agent._set_event_status(event.event_id, EventStatus.COMPLETED)
        assert event.status == EventStatus.COMPLETED.value
        assert mocked_agent._active_or_pending_events == []
        assert database.Event.get(status=EventStatus.COMPLETED.value) == event

    def test_from_unresponded_to_near(self, mocked_agent):
        event = EventFactory()
        mocked_agent._set_event_status(event.event_id, EventStatus.NEAR)
        assert event.status == EventStatus.NEAR.value
        assert mocked_agent._active_or_pending_events == [event]
        assert database.Event.get(status=EventStatus.NEAR.value) == event

    def test_from_unresponded_to_far(self, mocked_agent):
        event = EventFactory()
        mocked_agent._set_event_status(event.event_id, EventStatus.FAR)
        assert event.status == EventStatus.FAR.value
        assert mocked_agent._active_or_pending_events == [event]
        assert database.Event.get(status=EventStatus.FAR.value) == event

    def test_from_far_To_active(self, mocked_agent):
        event = EventFactory(
            status=EventStatus.FAR.value, opt_type=OptType.OPT_IN.value
        )
        assert event.status == EventStatus.FAR.value
        mocked_agent._set_event_status(event.event_id, EventStatus.ACTIVE)
        assert mocked_agent._active_or_pending_events == [event]
        assert database.Event.get(status=EventStatus.ACTIVE.value) == event

    def test_opt_in_from_unresponded_to_active(self, mocked_agent):
        event = EventFactory(opt_type=OptType.OPT_IN.value)
        mocked_agent._set_event_status(event.event_id, EventStatus.ACTIVE)
        assert event.status == EventStatus.ACTIVE.value
        assert mocked_agent._active_or_pending_events == [event]
        assert database.Event.get(status=EventStatus.ACTIVE.value) == event

    def test_opt_out_not_transitioned_from_responded_to_active(self, mocked_agent):
        event = EventFactory(opt_type=OptType.OPT_OUT.value)
        with pytest.raises(exceptions.InvalidStatusException):
            mocked_agent._set_event_status(event.event_id, EventStatus.ACTIVE)
            assert event.status != EventStatus.ACTIVE.value
            assert mocked_agent._active_or_pending_events == []

    def test_no_opt_not_transitioned_from_unresponded_to_active(self, mocked_agent):
        event = EventFactory(opt_type=OptType.NONE.value)
        with pytest.raises(exceptions.InvalidStatusException):
            mocked_agent._set_event_status(event.event_id, EventStatus.ACTIVE)
            assert event.status != EventStatus.ACTIVE.value
            assert mocked_agent._active_or_pending_events == []

    def test_from_active_to_completed(self, mocked_agent):
        event = EventFactory(status=EventStatus.ACTIVE.value)
        assert event.status == EventStatus.ACTIVE.value
        mocked_agent._set_event_status(event.event_id, EventStatus.COMPLETED)
        assert event.status == EventStatus.COMPLETED.value
        assert mocked_agent._active_or_pending_events == []
        assert database.Event.get(status=EventStatus.COMPLETED.value) == event

    def test_invalid_status_string(self, mocked_agent):
        event = EventFactory()
        status_before = event.status
        with pytest.raises(exceptions.InvalidStatusException):
            mocked_agent._set_event_status(event.event_id, "invalid thing!")
            assert event.status == status_before

    def test_invalid_status_enum(self, mocked_agent):
        event = EventFactory()
        status_before = event.status
        with pytest.raises(exceptions.InvalidStatusException):
            mocked_agent._set_event_status(event.event_id, ReportStatus.ACTIVE)
            assert event.status == status_before


class TestPublicEventsMethods:
    def test_active_or_pending_events(self, mocked_agent):
        event = EventFactory(status=EventStatus.ACTIVE.value)
        event2 = EventFactory(status=EventStatus.CANCELED.value)
        event3 = EventFactory()

        assert len(mocked_agent.active_events) == 1
        assert mocked_agent.active_events[0] == event.to_dict()

        assert len(mocked_agent._active_or_pending_events) == 2
        assert mocked_agent.active_or_pending_events[0] == event.to_dict()
        assert mocked_agent.active_or_pending_events[1] == event3.to_dict()


class TestProcessEvent:
    def test_active_event_opted_in_in_active_events(self, mocked_agent):
        event = EventFactory(
            start_time=datetime.utcnow() - timedelta(minutes=5),
            end_time=datetime.utcnow() + timedelta(hours=1),
            status=EventStatus.ACTIVE.value,
            opt_type=OptType.OPT_IN.value,
        )
        mocked_agent.process_event(event)
        assert event.status == EventStatus.ACTIVE.value
        assert event in mocked_agent._active_events
        assert mocked_agent.is_event_in_progress

    def test_active_event_opted_out_not_active_events(self, mocked_agent):
        """ An event shouldn't be able to get into this situation"""
        event = EventFactory(
            start_time=datetime.utcnow() - timedelta(minutes=5),
            end_time=datetime.utcnow() + timedelta(hours=1),
            status=EventStatus.ACTIVE.value,
            opt_type=OptType.OPT_OUT.value,
        )
        mocked_agent.process_event(event)
        assert event.status == EventStatus.ACTIVE.value
        assert event not in mocked_agent._active_events

    def test_event_completed_if_over(self, mocked_agent):
        event = EventFactory(
            start_time=datetime.utcnow() - timedelta(hours=1),
            end_time=datetime.utcnow() - timedelta(minutes=5),
            status=EventStatus.ACTIVE.value,
        )
        mocked_agent.process_event(event)
        assert event.status == EventStatus.COMPLETED.value
        assert event not in mocked_agent._active_events

    def test_unresponded_event_activated_if_started_and_opted_in(self, mocked_agent):
        event = EventFactory(
            start_time=datetime.utcnow() - timedelta(minutes=5),
            end_time=datetime.utcnow() + timedelta(hours=1),
            status=EventStatus.UNRESPONDED.value,
            opt_type=OptType.OPT_IN.value,
        )
        mocked_agent.process_event(event)
        assert event.status == EventStatus.ACTIVE.value
        assert event in mocked_agent._active_events
        assert mocked_agent.is_event_in_progress is True

    def test_unresponded_event_not_activated_if_started_and_opted_out(
        self, mocked_agent
    ):
        event = EventFactory(
            start_time=datetime.utcnow() - timedelta(minutes=5),
            end_time=datetime.utcnow() + timedelta(hours=1),
            status=EventStatus.UNRESPONDED.value,
            opt_type=OptType.OPT_OUT.value,
        )
        mocked_agent.process_event(event)
        assert event.status == EventStatus.UNRESPONDED.value
        assert event not in mocked_agent._active_events

    def test_unresponded_event_not_activated_if_started_and_no_opt_type(
        self, mocked_agent
    ):
        event = EventFactory(
            start_time=datetime.utcnow() - timedelta(minutes=5),
            end_time=datetime.utcnow() + timedelta(hours=1),
            status=EventStatus.UNRESPONDED.value,
            opt_type=OptType.NONE.value,
        )
        mocked_agent.process_event(event)
        assert event.status == EventStatus.UNRESPONDED.value
        assert event not in mocked_agent._active_events

    def test_active_event_opted_in_immediately_if_active_and_no_opt_type(
        self, mocked_agent
    ):
        event = EventFactory(
            start_time=datetime.utcnow() - timedelta(minutes=5),
            end_time=datetime.utcnow() + timedelta(hours=1),
            status=EventStatus.ACTIVE.value,
            opt_type=OptType.NONE.value,
        )
        mocked_agent.process_event(event)
        assert event.status == EventStatus.ACTIVE.value
        assert event in mocked_agent._active_events
        assert mocked_agent.is_event_in_progress is True


class TestWithTimeMovingOn:
    def test_event_opted_in_becomes_active(self, mocked_agent):
        initial_datetime = datetime(year=2019, month=1, day=1, hour=10, minute=0)
        with travel(initial_datetime) as traveller:
            event = EventFactory(
                start_time=initial_datetime + timedelta(minutes=5),
                end_time=datetime.utcnow() + timedelta(hours=1),
                created=initial_datetime,
                status=EventStatus.UNRESPONDED.value,
                opt_type=OptType.OPT_IN.value,
            )
            mocked_agent.process_event(event)
            assert event.status == EventStatus.UNRESPONDED.value
            traveller.shift(delta=timedelta(minutes=5))
            mocked_agent.process_event(event)
            assert event.status == EventStatus.ACTIVE.value

    def test_event_far_opted_in_becomes_active(self, mocked_agent):

        initial_datetime = datetime(year=2019, month=1, day=1, hour=10, minute=0)
        with travel(initial_datetime) as traveller:
            # assert traveller() == initial_datetime
            event = EventFactory(
                start_time=initial_datetime + timedelta(minutes=5),
                end_time=datetime.utcnow() + timedelta(hours=1),
                created=initial_datetime,
                status=EventStatus.FAR.value,
                opt_type=OptType.OPT_IN.value,
            )

            mocked_agent.process_event(event)
            assert event.status == EventStatus.FAR.value
            traveller.shift(delta=timedelta(minutes=5))
            mocked_agent.process_event(event)
            assert event.status == EventStatus.ACTIVE.value

    def test_event_opted_out_doesnt_become_active(self, mocked_agent):
        initial_datetime = datetime(year=2019, month=1, day=1, hour=10, minute=0)
        with travel(initial_datetime) as traveller:
            event = EventFactory(
                start_time=initial_datetime + timedelta(minutes=5),
                end_time=datetime.utcnow() + timedelta(hours=1),
                created=initial_datetime,
                status=EventStatus.UNRESPONDED.value,
                opt_type=OptType.OPT_OUT.value,
            )
            mocked_agent.process_event(event)
            assert event.status == EventStatus.UNRESPONDED.value
            traveller.shift(delta=timedelta(minutes=5))
            mocked_agent.process_event(event)
            assert event.status == EventStatus.UNRESPONDED.value


class TestDefaultForceOptIn:
    def test_event_opted_out_doesnt_become_active_before_force_opt_timeout_secs(
        self, mocked_agent
    ):
        initial_datetime = datetime(year=2019, month=1, day=1, hour=10, minute=0)
        with travel(initial_datetime) as traveller:
            event = EventFactory(
                start_time=initial_datetime + timedelta(minutes=5),
                end_time=datetime.utcnow() + timedelta(hours=1),
                created=initial_datetime,
                status=EventStatus.UNRESPONDED.value,
                opt_type=OptType.OPT_OUT.value,
            )
            mocked_agent.process_event(event)
            assert event.status == EventStatus.UNRESPONDED.value
            assert event.opt_type == OptType.OPT_OUT.value
            traveller.shift(
                delta=timedelta(seconds=(mocked_agent.opt_timeout_secs - 5))
            )
            mocked_agent.process_event(event)
            assert event.opt_type == OptType.OPT_OUT.value
            assert event.status == EventStatus.UNRESPONDED.value

    def test_event_opted_out_doesnt_become_opted_in_after_force_opt_timeout_secs(
        self, mocked_agent
    ):
        initial_datetime = datetime(year=2019, month=1, day=1, hour=10, minute=0)
        with travel(initial_datetime) as traveller:
            event = EventFactory(
                start_time=initial_datetime + timedelta(minutes=5),
                end_time=datetime.utcnow() + timedelta(hours=1),
                created=initial_datetime,
                status=EventStatus.UNRESPONDED.value,
                opt_type=OptType.OPT_OUT.value,
            )
            mocked_agent.process_event(event)
            assert event.status == EventStatus.UNRESPONDED.value
            assert event.opt_type == OptType.OPT_OUT.value
            traveller.shift(
                delta=timedelta(seconds=(mocked_agent.opt_timeout_secs + 5))
            )
            mocked_agent.process_event(event)
            assert event.opt_type == OptType.OPT_OUT.value

    def test_event_with_no_opt_type_doesnt_change_opt_type_before_force_opt_timeout_secs(
        self, mocked_agent
    ):
        initial_datetime = datetime(year=2019, month=1, day=1, hour=10, minute=0)
        with travel(initial_datetime) as traveller:
            event = EventFactory(
                start_time=initial_datetime + timedelta(minutes=5),
                end_time=datetime.utcnow() + timedelta(hours=1),
                created=initial_datetime,
                status=EventStatus.UNRESPONDED.value,
                opt_type=OptType.NONE.value,
            )
            mocked_agent.process_event(event)
            assert event.status == EventStatus.UNRESPONDED.value
            assert event.opt_type == OptType.NONE.value
            traveller.shift(
                delta=timedelta(seconds=(mocked_agent.opt_timeout_secs - 5))
            )
            mocked_agent.process_event(event)
            assert event.opt_type == OptType.NONE.value
            assert event.status == EventStatus.UNRESPONDED.value

    def test_event_with_no_opt_type_set_to_default_after_force_opt_timeout_secs(
        self, mocked_agent
    ):
        initial_datetime = datetime(year=2019, month=1, day=1, hour=10, minute=0)
        with travel(initial_datetime) as traveller:
            event = EventFactory(
                start_time=initial_datetime + timedelta(minutes=5),
                end_time=datetime.utcnow() + timedelta(hours=1),
                created=initial_datetime,
                status=EventStatus.UNRESPONDED.value,
                opt_type=OptType.NONE.value,
            )
            mocked_agent.process_event(event)
            assert event.status == EventStatus.UNRESPONDED.value
            assert event.opt_type == OptType.NONE.value
            traveller.shift(
                delta=timedelta(seconds=(mocked_agent.opt_timeout_secs + 5))
            )
            mocked_agent.process_event(event)
            assert event.opt_type == mocked_agent.opt_default_decision



from voluptuous import Schema

internal_schema = Schema(
    {
        "report_name_metadata": str,
        "report_interval_secs_default": int,
        "telemetry_parameters": dict,
    },
    required=True,
)
REPORT_PARAMETER_SCHEMA = Schema({str: internal_schema})



# Copyright 2017, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}
import json
import logging
import random
import sys
from collections import namedtuple
from datetime import datetime
from datetime import timedelta
from io import StringIO
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple

import isodate
import lxml.etree as etree_
import pem
import requests
import signxml
from pony import orm
from requests.exceptions import ConnectionError

from . import builders
from . import database
from . import enums
from . import exceptions
from . import extractors
from . import models
from . import oadr_20b
from . import response_codes
from .utils import get_aware_utc_now
from .validation_schemas import REPORT_PARAMETER_SCHEMA

_LOGGER = logging.getLogger(__name__)

ENDPOINT_BASE = "/OpenADR2/Simple/2.0b/"
EIEVENT = ENDPOINT_BASE + "EiEvent"
EIREPORT = ENDPOINT_BASE + "EiReport"
EIREGISTERPARTY = ENDPOINT_BASE + "EiRegisterParty"
POLL = ENDPOINT_BASE + "OadrPoll"

Endpoint = namedtuple("Endpoint", ["url", "callback"])
OPENADR_ENDPOINTS = {
    "EiEvent": Endpoint(url=EIEVENT, callback="push_request"),
    "EiReport": Endpoint(url=EIREPORT, callback="push_request"),
    "EiRegisterParty": Endpoint(url=EIREGISTERPARTY, callback="push_request"),
}


MIN_LOOP_FREQUENCY_SECS = 5
DEFAULT_POLL_FREQUENCY_SECS = 15
DEFAULT_REPORT_INTERVAL_SECS = 15
DEFAULT_OPT_TIMEOUT_SECS = 30 * 60


class OpenADRVenAgent:
    """
        OpenADR (Automated Demand Response) is a standard for alerting and responding
        to the need to adjust electric power consumption in response to fluctuations
        in grid demand.

        OpenADR communications are conducted between Virtual Top Nodes (VTNs) and Virtual End Nodes (VENs).
        In this implementation, a this agent is a VEN, implementing EiEvent and EiReport services
        in conformance with a subset of the OpenADR 2.0b specification.

        The VEN receives VTN requests via the web service.

        The VTN can 'call an event', indicating that a load-shed event should occur.
        The VEN responds with an 'optIn' acknowledgment.

        Events:
            The VEN agent maintains a persistent record of DR events.
            These events are stored in a sqlite database

        Reporting:
            The VEN agent configuration defines telemetry values (data points) to be reported to the VTN.
            The VEN agent maintains a persistent record of reportable/reported telemetry values over time.

        Supported requests/responses in the OpenADR VTN interface:
            VTN:
                oadrDistributeEvent (needed for event cancellation)
                oadrResponse
                oadrRegisteredReport
                oadrCreateReport
                oadrUpdatedReport
                oadrCancelReport
                oadrCreatedPartyRegistration
            VEN:
                oadrPoll
                oadrRequestEvent
                oadrCreatedEvent
                oadrResponse
                oadrRegisterReport
                oadrCreatedReport
                oadrUpdateReport
                oadrCanceledReport
                oadrCreatePartyRegistration
                oadrQueryRegistration


        Event workflow (see OpenADR Profile Specification section 8.1)...

        Event poll / creation:
            (VEN) oadrPoll
            (VTN) oadrDistributeEvent (all events are included; one oadrEvent element per event)
            (VEN) oadrCreatedEvent with optIn/optOut (if events had oadrResponseRequired)
                    If "always", an oadrCreatedEvent must be sent for each event.
                    If "never", it was a "broadcast" event -- never create an event in response.
                    Otherwise, respond if event state (eventID, modificationNumber) has changed.
            (VTN) oadrResponse

        Event change:
            (VEN) oadrCreatedEvent (sent if the optIn/optOut status has changed)
            (VTN) oadrResponse

        Sample oadrDistributeEvent use case from the OpenADR Program Guide:

            Event:
                Notification: Day before event
                Start Time: midnight
                Duration: 24 hours
                Randomization: None
                Ramp Up: None
                Recovery: None
                Number of signals: 2
                Signal Name: simple
                    Signal Type: level
                    Units: LevN/A
                    Number of intervals: equal TOU Tier change in 24 hours (2 - 6)
                    Interval Duration(s): TOU tier active time frame (i.e. 6 hours)
                    Typical Interval Value(s): 0 - 4 mapped to TOU Tiers (0 - Cheapest Tier)
                    Signal Target: None
                Signal Name: ELECTRICITY_PRICE
                    Signal Type: price
                    Units: USD per Kwh
                    Number of intervals: equal TOU Tier changes in 24 hours (2 - 6)
                    Interval Duration(s): TOU tier active time frame (i.e. 6 hours)
                    Typical Interval Value(s): $0.10 to $1.00 (current tier rate)
                    Signal Target: None
                Event Targets: venID_1234
                Priority: 1
                VEN Response Required: always
                VEN Expected Response: optIn
            Reports:
                None

        Report workflow (see OpenADR Profile Specification section 8.3)...

        Report registration interaction:
            (VEN) oadrRegisterReport (METADATA report)
                VEN sends its reporting capabilities to VTN.
                Each report, identified by a reportSpecifierID, is described as elements and attributes.
            (VTN) oadrRegisteredReport (with optional oadrReportRequests)
                VTN acknowledges that capabilities have been registered.
                VTN optionally requests one or more reports by reportSpecifierID.
                Even if reports were previously requested, they should be requested again at this point.
            (VEN) oadrCreatedReport (if report requested)
                VEN acknowledges that it has received the report request and is generating the report.
                If any reports were pending delivery, they are included in the payload.
            (VTN) oadrResponse
                Why??

        Report creation interaction:
            (VTN) oadrCreateReport
                See above - this is like the "request" portion of oadrRegisteredReport
            (VEN) oadrCreatedReport
                See above.

        Report update interaction - this is the actual report:
            (VEN) oadrUpdateReport (report with reportRequestID and reportSpecifierID)
                Send a report update containing actual data values
            (VTN) oadrUpdatedReport (optional oadrCancelReport)
                Acknowledge report receipt, and optionally cancel the report

        Report cancellation:
            (VTN) oadrCancelReport (reportRequestID)
                This can be sent to cancel a report that is in progress.
                It should also be sent if the VEN keeps sending oadrUpdateReport
                    after an oadrUpdatedReport cancellation.
                If reportToFollow = True, the VEN is expected to send one final additional report.
            (VEN) oadrCanceledReport
                Acknowledge the cancellation.
                If any reports were pending delivery, they are included in the payload.

        Key elements in the METADATA payload:
            reportSpecifierID: Report identifier, used by subsequent oadrCreateReport requests
            rid: Data point identifier
                This VEN reports only two data points: baselinePower, actualPower
            Duration: the amount of time that data can be collected
            SamplingRate.oadrMinPeriod: maximum sampling frequency
            SamplingRate.oadrMaxPeriod: minimum sampling frequency
            SamplingRate.onChange: whether or not data is sampled as it changes

        For an oadrCreateReport example from the OpenADR Program Guide, see test/xml/sample_oadrCreateReport.xml.

    """

    EiReports: List[models.EiReport] = []
    EiTelemetryValuess: List[models.EiTelemetryValues] = []

    _last_poll: Optional[datetime] = None

    reports_registered: bool = False

    def __init__(
        self,
        *,
        ven_id: str,
        vtn_id: str,
        vtn_address: str,
        client_pem_bundle: str,
        vtn_ca_cert: str,
        db_filepath: Optional[str] = None,
        log_xml: bool = False,
        opt_default_decision: str = enums.OptType.OPT_IN.value,
        opt_timeout_secs: int = DEFAULT_OPT_TIMEOUT_SECS,
        poll_interval_secs: int = DEFAULT_POLL_FREQUENCY_SECS,
        report_parameters={},
        security_level: str = "standard",
    ):
        """
        Initialize the agent.


        :param ven_id: OpenADR ID of this virtual end node. Identifies this VEN to the VTN.
        :param vtn_id: ID of the VTN with which this VEN will communicate
        :param vtn_address: URL of the VTN
        :param client_pem_bundle: The pem bundle (private key + certificate concatenated) to present to the server for HTTPS mutual auth
        :param vtn_ca_cert: The path of the certificate from the CA that the client certs have been signed with.
        :param db_filepath: The path to the SQLite database file.  If not provided, will be kept in-memory.
        :param log_xml: Whether to write inbound/outbound XML to the agent's log
        :param opt_default_decision: What opt-in/opt-out choice to make by default
        :param opt_timeout_secs: How long to wait before making a default opt-in/opt-out decision
        :param poll_interval_secs: How often the VEN should send an OadrPoll to the VTN
        :param security_level: If 'high', the VTN and VEN use a third-party signing authority to sign and authenticate each request. Default is 'standard' (XML payloads do not contain Signature elements)
        :param report_parameters: A dictionary of definitions of reporting/telemetry parameters

        The ``report_parameters`` option should look like so:

        .. code-block:: python

            report_parameters={
                "reportSpecifierId": {
                    "report_interval_secs_default": 15,
                    "report_name": name,
                    "telemetry_parameters": { ... },  # Should be JSON-safe
                }
            }

        """
        self.ven_id = ven_id
        self.vtn_id = vtn_id
        self.vtn_address = vtn_address
        self.client_pem_bundle = client_pem_bundle
        self.vtn_ca_cert = vtn_ca_cert
        self.log_xml = log_xml
        self.opt_default_decision = opt_default_decision
        self.opt_timeout_secs = opt_timeout_secs
        self.poll_interval_secs = poll_interval_secs

        self.security_level = security_level

        try:
            REPORT_PARAMETER_SCHEMA(report_parameters)
            self.report_parameters = report_parameters
        except:
            _LOGGER.exception("Report parameters are incorrect")
            raise

        if self.poll_interval_secs < MIN_LOOP_FREQUENCY_SECS:
            raise ValueError(
                f"Poll interval is too frequent; use one higher than {MIN_LOOP_FREQUENCY_SECS}"
            )

        if "pytest" not in sys.modules:
            database.setup_db(db_filepath)

        # State variables for VTN request/response processing
        self.ven_online = "false"
        self.ven_manual_override = "false"

        _LOGGER.info("Configuration parameters:")
        _LOGGER.info(f"  ven_id = {ven_id}")
        _LOGGER.info(f"  vtn_id = {vtn_id}")
        _LOGGER.info(f"  vtn_address = {vtn_address}")
        _LOGGER.info(f"  client_pem_bundle = {client_pem_bundle}")
        _LOGGER.info(f"  vtn_ca_cert = {vtn_ca_cert}")
        _LOGGER.info(f"  log_xml = {log_xml}")
        _LOGGER.info(f"  opt_default_decision = {opt_default_decision}")
        _LOGGER.info(f"  opt_timeout_secs = {opt_timeout_secs}")
        _LOGGER.info(f"  poll_interval_secs = {poll_interval_secs}")
        _LOGGER.info(f"  report_parameters = {report_parameters}")
        _LOGGER.info(f"  security_level = {security_level}")

    def register(self, ven_name: str) -> dict:
        """
        Send a one-time registration request to the VTN using oadrCreatePartyRegistration

        This results in a dictionary with up to three keys.  It definitely will contain
        ``ven_id`` and ``vtn_id``.  It may contain ``poll_interval_secs``. These values
        should be stored somewhere safe, and passed to future initialisations of the
        ``OpenADRVenAgent``.

        :param ven_name: The name of this VEN, according to the OpenADR specification.
        """
        if not ven_name:
            raise ValueError("You need a ven_name to register")

        # VEN registration with the VTN server.
        # Register the VEN, obtaining the VEN ID. This is a one-time action.
        self.send_oadr_create_party_registration(ven_name)

        result = {
            "vtn_id": self.vtn_id,
            "ven_id": self.ven_id,
            "poll_interval_secs": self.poll_interval_secs,
        }

        # We do this to get around the fact that the server might have returned nothing
        # for the poll interval.  It's easier to set this value to None in the depths of
        # the server response handling code (see handle_oadr_created_party_registration)
        # and re-set it here to something sensible, than it is to restructure the code
        # so that it could e.g. return a value that might percolate up to this layer.
        # As such, it is a code smell.   XXX
        if not self.poll_interval_secs:
            self.poll_interval_secs = DEFAULT_POLL_FREQUENCY_SECS

        return result

    def request_events(self):
        """
        Make a request to the VTN for all current events.
        """
        self.send_oadr_request_event()

    def tick(self):
        """
        Perform the following tasks:
        - Polling the VTN server, if it is long enough since the last poll
        - Perform event management tasks:
          - force an optIn/optOut decision if too much time has elapsed.
          - transition event state when appropriate.
          - expire events that have become completed or canceled.
        - Perform report management tasks:
          - send telemetry to the VTN for any active report.
          - transition report state when appropriate.
          - expire reports that have become completed or canceled.

        The first time this function is called in an instance's lifetime, it will
        send an oadrRegisterReport request.

        This method should be called by your update method every clock tick.
        """
        if not self.reports_registered:
            # Send an initial report-registration request to the VTN.
            self.send_oadr_register_report()
            self.reports_registered = True

        # If it's been poll_interval_secs since the last poll request, issue a new one.
        if self._last_poll is None or (
            (get_aware_utc_now() - self._last_poll).total_seconds()
            > self.poll_interval_secs
        ):
            self.send_oadr_poll()

        for event in self._active_or_pending_events:
            self.process_event(event)

        for report in self.active_or_pending_reports:
            self.process_report(report.id)

    @orm.db_session
    def process_event(self, event: database.Event):
        """
        Perform periodic maintenance for an event.

        Transition its state when appropriate.
        """

        _LOGGER.debug(f"Processing event::: {event.event_id}")
        if event.is_active_or_pending:
            _LOGGER.debug(f"Event {event.event_id} is {event.status}")
            _LOGGER.debug(f"{event}")
            self._complete_event_if_over(event.event_id)
            self._force_opt_type(event.event_id)
            if event.status != enums.EventStatus.ACTIVE.value:
                self._activate_event_if_starts_now(event)

    @orm.db_session
    def _force_opt_type(self, event_id):
        _LOGGER.debug("Forcing Opt Type")
        _LOGGER.debug(f"opt types ")
        event = self.get_event_for_id(event_id)
        if event.opt_type != enums.OptType.NONE.value:
            _LOGGER.debug(f"opt type is not none {event.opt_type}")
            return
        if self.opt_default_decision not in enums.OptType.list():
            _LOGGER.debug("not in opt type list")
            return
        if event.status == enums.EventStatus.ACTIVE.value:
            event.opt_type = self.opt_default_decision
        _LOGGER.debug(f"opt timeout secs {self.opt_timeout_secs}")
        if not bool(self.opt_timeout_secs):
            event.opt_type = self.opt_default_decision
        if datetime.utcnow() >= event.created + timedelta(
            seconds=self.opt_timeout_secs
        ):
            event.opt_type = self.opt_default_decision

    @orm.db_session
    def _complete_event_if_over(self, event_id):
        event = self.get_event_for_id(event_id)
        now = datetime.utcnow()
        if not event.is_active_or_pending:
            return
        if event.end_time is None:
            return
        if now > event.end_time:
            _LOGGER.debug(
                "Setting event %s to status %s",
                event.event_id,
                enums.EventStatus.COMPLETED.value,
            )
            self._set_event_status(event.event_id, enums.EventStatus.COMPLETED)

    @orm.db_session
    def _activate_event_if_starts_now(self, event):
        now = datetime.utcnow()
        if now < event.start_time:
            return
        if event.opt_type != enums.OptType.OPT_IN.value:
            return
        _LOGGER.debug(
            "Setting event {} to status {}".format(
                event.event_id, enums.EventStatus.ACTIVE.value
            )
        )
        self._set_event_status(event.event_id, enums.EventStatus.ACTIVE)

    @orm.db_session
    def process_report(self, rpt_id: int):
        """
        Perform periodic maintenance for a report.

        Send telemetry to the VTN if the report is active.
        Transition its state when appropriate.
        Expire it from the cache if it has become completed or canceled.
        """
        rpt = database.Report.get(id=rpt_id)
        _LOGGER.debug(f"Processing report named: {rpt.name}")

        def _activate_report():
            if not rpt.start_time:
                return
            if rpt.start_time < now and (rpt.end_time is None or now < rpt.end_time):
                _LOGGER.debug(
                    "Setting report {} to status {}".format(
                        rpt.request_id, enums.ReportStatus.ACTIVE.value
                    )
                )
                self.set_report_status(rpt, enums.ReportStatus.ACTIVE)

        if rpt.is_active_or_pending:
            now = datetime.utcnow()
            if rpt.status == enums.ReportStatus.ACTIVE.value:
                if rpt.end_time is None or rpt.end_time > now:
                    rpt_interval = (
                        rpt.interval_secs
                        if rpt.interval_secs is not None
                        else DEFAULT_REPORT_INTERVAL_SECS
                    )
                    if not rpt.last_report or datetime.utcnow() > (
                        rpt.last_report + timedelta(seconds=rpt_interval)
                    ):
                        self.send_oadr_update_report(rpt.id)
                        if rpt_interval == 0:
                            # OADR rule 324: If rpt_interval == 0 it's a one-time report, so set status to COMPLETED.
                            rpt.status = enums.ReportStatus.COMPLETED.value
                else:
                    _LOGGER.debug(
                        "Setting report {} to status {}".format(
                            rpt.request_id, enums.ReportStatus.COMPLETED.value
                        )
                    )
                    self.set_report_status(rpt, enums.ReportStatus.COMPLETED)
            else:
                _activate_report()

    # ***************** Methods for Servicing VTN Requests ********************

    # def push_request(self, env, request):
    #     """Callback. The VTN pushed an http request. Service it."""
    #     _LOGGER.debug("Servicing a VTN push request")
    #     # self.core.spawn(self.service_vtn_request, request)    # **HA**
    #     self.service_vtn_request(request)
    #     # Return an empty response.
    #     return [response_codes.HTTP_STATUS_CODES[204], "", [("Content-Length", "0")]]

    RequestHandler = namedtuple(
        "RequestHandler", ["func", "ei_response", "id_location"]
    )

    RESP_ALWAYS = "always"
    RESP_OPTIONAL = "optional"
    RESP_NEVER = "never"

    LOCATION_ROOT = "root"
    LOCATION_RESPONSE = "ei_response"
    LOCATION_NONE = "none"

    REQUEST_HANDLERS = {
        "oadrDistributeEvent": RequestHandler(
            func="handle_oadr_distribute_event",
            ei_response=RESP_OPTIONAL,
            id_location=LOCATION_ROOT,
        ),
        "oadrRegisterReport": RequestHandler(
            func="handle_oadr_register_report",
            ei_response=RESP_NEVER,
            id_location=LOCATION_NONE,
        ),
        "oadrRegisteredReport": RequestHandler(
            func="handle_oadr_registered_report",
            ei_response=RESP_ALWAYS,
            id_location=LOCATION_RESPONSE,
        ),
        "oadrCreateReport": RequestHandler(
            func="handle_oadr_create_report",
            ei_response=RESP_NEVER,
            id_location=LOCATION_NONE,
        ),
        "oadrUpdatedReport": RequestHandler(
            func="handle_oadr_updated_report",
            ei_response=RESP_ALWAYS,
            id_location=LOCATION_RESPONSE,
        ),
        "oadrCancelReport": RequestHandler(
            func="handle_oadr_cancel_report",
            ei_response=RESP_NEVER,
            id_location=LOCATION_ROOT,
        ),
        "oadrResponse": RequestHandler(
            func="handle_oadr_response",
            ei_response=RESP_ALWAYS,
            id_location=LOCATION_RESPONSE,
        ),
        "oadrCreatedPartyRegistration": RequestHandler(
            func="handle_oadr_created_party_registration",
            ei_response=RESP_ALWAYS,
            id_location=LOCATION_RESPONSE,
        ),
    }

    @staticmethod
    def _find_request_handler(handlers: Dict[str, RequestHandler], signed_object):
        """
        Given a SignedObject from the VTN, return the request handler that tells us
        how to deal with it.
        """
        found = [
            payload_name
            for payload_name in handlers.keys()
            if getattr(signed_object, payload_name)
        ]

        if len(found) == 1:
            payload_name = found[0]
            return payload_name, handlers[payload_name]
        elif len(found) == 0:
            raise exceptions.OpenADRInterfaceException(
                f"Did not recognise payload.  Supported types are {handlers.keys()}",
                response_codes.OADR_NOT_RECOGNISED,
            )
        else:
            raise exceptions.OpenADRInterfaceException(
                f"Too many signedObject elements ({','.join(found)})",
                response_codes.OADR_INVALID_DATA,
            )

    @staticmethod
    def _get_signed_object(request):
        payload = oadr_20b.parseString(request, silence=True)
        signed_object = payload.oadrSignedObject
        if signed_object is None:
            raise exceptions.OpenADRInterfaceException(
                "No SignedObject in payload", response_codes.OADR_BAD_DATA
            )
        return signed_object

    def _check_ei_response(self, request_object, check_type, id_location) -> str:
        """
        An eiResponse can appear in multiple kinds of VTN requests.

        If an eiResponse has been received, check for a '200' (OK) response code.
        If any other code is received, the VTN is reporting an error -- log it and raise an exception.

        :param request_object: The VTN's request object..
        :return: The request ID
        """
        if check_type == self.RESP_ALWAYS or (
            check_type == self.RESP_OPTIONAL and getattr(request_object, "eiResponse")
        ):
            if (
                request_object.eiResponse.responseCode
                != response_codes.OADR_VALID_RESPONSE
            ):
                raise exceptions.OpenADRInternalException(
                    f"Error response from VTN: {request_object.eiResponse.responseCode}"
                    f" {request_object.eiResponse.responseDescription}"
                )

        if id_location == self.LOCATION_NONE:
            return ""
        elif id_location == self.LOCATION_ROOT:
            # OADR rule 41: There are two requestID fields in oadrDistributeEvent.
            # The field that must be populated with a requestID is located
            # at oadrDistributeEvent:requestID.
            return request_object.requestID
        else:
            return request_object.eiResponse.requestID

    def service_vtn_request(self, request: str):
        """
        Handle a payload received from the VTN.

        :param request: The request's XML payload (as a string)
        """
        if self.log_xml:
            from bs4 import BeautifulSoup

            bs = BeautifulSoup(request, "xml")
            _LOGGER.debug(f"VTN payload:\n{bs.prettify()}")

        if self.security_level == "high":
            # At high security, the request is accompanied by a Signature.
            # The VEN should use a certificate authority to validate and decode the request.
            raise NotImplementedError("Security level 'high' is not implemented")

        #  We set this here so that it is defined before the exception handling blocks
        # which rely on it being set to *something*.
        request_id = ""

        try:
            signed_object = self._get_signed_object(request)
            payload_name, handler = self._find_request_handler(
                self.REQUEST_HANDLERS, signed_object
            )
            _LOGGER.debug(f"VTN request: {payload_name} -> {handler.func}")

            # e.g. signed_object.oadrDistributeEvent
            request_object = getattr(signed_object, payload_name)
            request_method = getattr(self, handler.func)

            request_id = self._check_ei_response(
                request_object, handler.ei_response, handler.id_location
            )
            request_method(request_id, request_object)
            if request_object.__class__.__name__ != "oadrResponseType":
                # A non-default response was received from the VTN. Issue a followup poll request.
                self.send_oadr_poll()

        except exceptions.OpenADRInternalException as err:
            _LOGGER.exception(
                f"Internal error when handling VTN request ({err.message})"
            )
        except exceptions.OpenADRInterfaceException as err:
            _LOGGER.warning(
                f"Error processing VTN request ({err.message})", exc_info=True
            )
            # OADR rule 48: Log the validation failure, send an oadrResponse.eiResponse with an error code.
            self.send_oadr_response(
                err.message, err.error_code or response_codes.OADR_BAD_DATA, request_id
            )
        except Exception as err:
            _LOGGER.exception("Unexpected error handling VTN request")
            self.send_oadr_response(str(err), response_codes.OADR_BAD_DATA, request_id)
            raise

    # ***************** Handle Requests from the VTN to the VEN ********************

    def handle_oadr_created_party_registration(
        self, request_id, oadr_created_party_registration
    ):
        """
        The VTN has responded to oadrCreatePartyRegistration
        by sending oadrCreatedPartyRegistration.
        """
        registration = extractors.decode_oadr_created_party_registration(
            oadr_created_party_registration
        )

        self.vtn_id = registration.vtn_id
        self.ven_id = registration.ven_id
        self.poll_interval_secs = registration.poll_interval_secs

    def handle_oadr_distribute_event(self, request_id, oadr_distribute_event):
        """
        The VTN has responded to an oadrPoll by sending an oadrDistributeEvent.

        Create or update an event, then respond with oadrCreatedEvent.

        For sample XML, see test/xml/sample_oadrDistributeEvent.xml.
        """

        vtn_id = oadr_distribute_event.vtnID
        if vtn_id is not None and vtn_id != self.vtn_id:
            raise exceptions.OpenADRInterfaceException(
                "vtnID failed to match agent config: {}".format(vtn_id),
                response_codes.OADR_BAD_DATA,
            )

        oadr_event_list = oadr_distribute_event.oadrEvent
        if len(oadr_event_list) == 0:
            raise exceptions.OpenADRInternalException(
                "oadrDistributeEvent received with no events"
            )

        oadr_event_ids = []
        for oadr_event in oadr_event_list:
            try:
                event = self.handle_oadr_event(request_id, oadr_event)
                if event:
                    oadr_event_ids.append(event.event_id)
            except exceptions.OpenADRInterfaceException as err:
                # OADR rule 19: If a VTN message contains a mix of valid and invalid events,
                # respond to the valid ones. Don't reject the entire message due to invalid events.
                # OADR rule 48: Log the validation failure and send the error code in oadrCreatedEvent.eventResponse.
                # (The oadrCreatedEvent's eiResponse should contain a 200 -- normal -- status code.)
                _LOGGER.warning("Event error: {}".format(err), exc_info=True)

                # Construct a temporary EIEvent to hold data that will be reported in the error return.
                if oadr_event.eiEvent and oadr_event.eiEvent.eventDescriptor:
                    event_id = oadr_event.eiEvent.eventDescriptor.eventID
                    modification_number = (
                        oadr_event.eiEvent.eventDescriptor.modificationNumber
                    )
                else:
                    event_id = None
                    modification_number = None
                with orm.db_session:
                    error_event = database.Event(
                        event_id=event_id,
                        request_id=request_id,
                        modification_number=modification_number,
                    )
                self.send_oadr_created_event(
                    error_event,
                    error_code=err.error_code or response_codes.OADR_BAD_DATA,
                    error_message=err.message,
                )

            except Exception as err:
                _LOGGER.warning(
                    "Unanticipated error during event processing: {}".format(err),
                    exc_info=True,
                )
                self.send_oadr_response(
                    str(err), response_codes.OADR_BAD_DATA, request_id
                )

        self._cancel_events_not_in_recent_event_list(oadr_event_ids)

    @orm.db_session
    def _cancel_events_not_in_recent_event_list(self, oadr_event_ids):
        """
        Implied cancel:

        OADR rule 61: If the VTN request omitted an active event, cancel it.

        Also, think about whether to alert the VTN about this cancellation by sending it an oadrCreatedEvent.
        """
        stored_events = database.Event.select()[:]
        for event in stored_events:
            if event.event_id not in oadr_event_ids:
                _LOGGER.debug(
                    "Event ID {} not in distributeEvent: canceling it.".format(
                        event.event_id
                    )
                )
                self.handle_event_cancellation(event, "never")

    @orm.db_session
    def handle_oadr_event(
        self, request_id, oadr_event: etree_.Element
    ) -> Optional[database.Event]:
        """
        An oadrEvent was received, usually as part of an oadrDistributeEvent.
        Handle the event creation/update.

        Respond with oadrCreatedEvent.

        For sample XML, see test/xml/sample_oadrDistributeEvent.xml.

        :return: the event that was created or updated
        """

        def _send_response_if_required(event):
            """OADR Rule 12
                The VEN MUST respond with an oadrCreatedEvent to an event in oadrDis-
                tributeEvent based upon the value in each event’s oadrResponseRequired
                element as follows:
                "always" – The VEN MUST respond to the event with an oadrCreatedEvent
                eventResponse. This includes unchanged, new, changed, and canceled
                events.
                "never" – The VEN MUST NOT respond to the event with a oadrCreatedEvent
                eventResponse
                Note that oadrCreatedEvent event responses SHOULD be returned in one
                message, but MAY be returned in separate messages.
            """
            if response_required == "always":
                if event is None:
                    return
                # OADR rule 12, 62: Send an oadrCreatedEvent if response_required == 'always'.
                # OADR rule 12, 62: If response_required == 'never', do not send an oadrCreatedEvent
                _LOGGER.warning("Sending Response for:")

                _LOGGER.warning(
                    f"Event ID: {event.event_id}, Event Status: {event.status}"
                )
                self.send_oadr_created_event(event)

        def _calculate_start_offset(event_data):
            # OADR rule 30: Randomize start_time and end_time if start_after is provided.
            if event_data.start_after:
                seconds = event_data.start_after.seconds * random.random()
                return timedelta(seconds=seconds)
            else:
                return timedelta(seconds=0)

        def _calculate_times(event_data, offset) -> Tuple[datetime, Optional[datetime]]:
            start_time = event_data.official_start + offset

            # An interval with 0 duration has no defined endTime and remains active
            # until canceled.
            if event_data.duration.total_seconds() <= 0.0:
                end_time = None
            else:
                end_time = start_time + event_data.duration

            return start_time, end_time

        @orm.db_session
        def _create_pony_event(
            event_data: extractors.EiEventData, request_id
        ) -> database.Event:
            """Create an Event in the databse."""

            start_offset = _calculate_start_offset(event_data)
            start_time, end_time = _calculate_times(event_data, start_offset)

            pony_event = database.Event(
                event_id=event_data.event_id,
                request_id=request_id,
                status=event_data.status.value,
                modification_number=event_data.modification_number,
                priority=event_data.priority,
                official_start=event_data.official_start,
                duration=event_data.duration,
                start_after=event_data.start_after,
                start_offset=start_offset,
                start_time=start_time,
                end_time=end_time,
                signals=event_data.signals,
                test_event=event_data.test_event,
            )

            return pony_event

        @orm.db_session
        def _update_pony_event(
            new_event_data: extractors.EiEventData, existing_event_id, request_id
        ):
            event = database.Event.get(id=existing_event_id)
            if event.opt_type == enums.OptType.OPT_OUT:
                # TODO: refactor
                # Do nothing if we have opted out of this event
                return
            if new_event_data.status is enums.EventStatus.CANCELED:
                if event.status != enums.EventStatus.CANCELED.value:
                    # OADR rule 59: The event was just canceled. Process an event cancellation.
                    self.handle_event_cancellation(event, response_required)
                    return event

            if event.status == enums.EventStatus.CANCELED.value:
                if new_event_data.status is not enums.EventStatus.CANCELED:
                    self.handle_event_uncancellation(
                        event, new_event_data, response_required
                    )
                    return event

            # Only recalculate date-related values if the VTN has changed them.
            #
            # This is because otherwise, every time we re-recieve an event with a
            # start_after (randomisation) value we would re-randomise the start and end
            # times.  Maybe that would be finebut new_event_dataems edge-casey so we avoid it.
            if (
                event.official_start != new_event_data.official_start
                or event.duration != new_event_data.duration
                or event.start_after != new_event_data.start_after
            ):
                event.start_offset = _calculate_start_offset(new_event_data)
                event.start_time, event.end_time = _calculate_times(
                    new_event_data, event.start_offset
                )

            event.event_id = new_event_data.event_id
            event.request_id = request_id
            event.priority = new_event_data.priority
            event.signals = new_event_data.signals
            event.modification_number = new_event_data.modification_number
            event.test_event = new_event_data.test_event
            event.official_start = new_event_data.official_start
            event.duration = new_event_data.duration
            event.start_after = new_event_data.start_after

            # XXX Is there a reason we don't set event.status to the VTN's provided status?
            return event

        # Create a temporary EiEvent, constructed from the OadrDistributeEventType.
        ei_event = oadr_event.eiEvent
        response_required = oadr_event.oadrResponseRequired

        if (
            ei_event.eiTarget
            and ei_event.eiTarget.venID
            and self.ven_id not in ei_event.eiTarget.venID
        ):
            # Rule 22: If an eiTarget is furnished, handle the event only if this venID
            # is in the target list.
            return None
        else:
            event_data = extractors.decode_ei_event(ei_event)
            existing_event = database.Event.get(event_id=event_data.event_id)
            if existing_event:

                # if event exists already, check that the modification number has been incremented
                # if it has, update the event

                if event_data.modification_number > existing_event.modification_number:
                    event = _update_pony_event(
                        event_data, existing_event.id, request_id
                    )
                    _send_response_if_required(event)
                elif (
                    event_data.modification_number == existing_event.modification_number
                ):
                    event = existing_event
                    _send_response_if_required(event)

                else:
                    _LOGGER.debug(
                        f"Out-of-order modification number: {event_data.modification_number}"
                    )
                    # OADR rule 58: Respond with error code 450.
                    raise exceptions.OpenADRInterfaceException(
                        "Invalid modification number (too low)",
                        response_codes.OADR_MOD_NUMBER_OUT_OF_ORDER,
                    )

            else:
                # create a new event
                event = _create_pony_event(event_data, request_id)
                _send_response_if_required(event)

            return event

    @orm.db_session
    def handle_event_cancellation(self, event: database.Event, response_required: str):
        """
        An event was canceled by the VTN. Update local state and publish the news.

        :param event: The event that was cancelled.
        :param response_required: Does the VTN expect a response?  Can be "always" or "never".
        """
        if bool(event.start_after):
            raise NotImplementedError(
                "Event cancellation with start_after not supported - see OADR rule 65"
            )
            """ OADR rule 65: If the event has a startAfter value,
                schedule cancellation for a random future time between now and (now + startAfter).
            """

        else:
            event.status = enums.EventStatus.CANCELED.value
            if response_required != "never":
                # OADR rule 36: If response_required != never, confirm cancellation with optType = optIn.
                event.optType = enums.OptType.OPT_IN.value

    @orm.db_session
    def handle_event_uncancellation(
        self, event: database.Event, new_event_data, response_required: str
    ):
        """
        A previously cancelled event was uncancelled (server side opt out/in).
        Update local state and publish the news.

        :param event: The event that was uncancelled.
        :param response_required: Does the VTN expect a response?  Can be "always" or "never".
        """
        event.status = new_event_data.status.value
        if response_required != "never":
            # OADR rule 36: If response_required != never, confirm cancellation with optType = optIn.
            event.optType = enums.OptType.OPT_IN.value

    def handle_oadr_register_report(self, request_id, oadr_register_report):
        """
        The VTN is sending METADATA, registering the reports that it can send to the VEN.

        Send no response -- the VEN doesn't want any of the VTN's crumby reports.

        :param request: The VTN's request (unused)
        """
        # OADR rule 301: Sent when the VTN wakes up.
        pass

    def handle_oadr_registered_report(self, request_id, oadr_registered_report):
        """
        The VTN acknowledged receipt of the METADATA in oadrRegisterReport.

        If the VTN requested any reports (by specifier ID), create them.
        Send an oadrCreatedReport acknowledgment for each request.

        :param oadr_registered_report: the oadrRegisteredReportType element
        """
        _LOGGER.debug(
            "Handling oadrRegisteredReport from VTN: {}".format(
                str(oadr_registered_report)
            )
        )
        self.create_or_update_reports(
            oadr_registered_report.oadrReportRequest, request_id
        )

    def handle_oadr_create_report(self, request_id, oadr_create_report):
        """
        Handle an oadrCreateReport request from the VTN.

        The request could have arrived in response to a poll,
        or it could have been part of an oadrRegisteredReport response.

        Create a report for each oadrReportRequest in the list, sending an oadrCreatedReport in response.

        :param oadr_create_report: the oadrCreateReport element
        """

        self.create_or_update_reports(oadr_create_report.oadrReportRequest, None)

    def handle_oadr_updated_report(self, request_id, oadr_updated_report):
        """
        The VTN acknowledged receipt of an oadrUpdatedReport, and may have sent
        a report cancellation.

        Check for report cancellation, and cancel the report if necessary. No need to
        send a response to the VTN.

        :param oadr_updated_report: the oadrUpdatedReport element
        """
        oadr_cancel_report = oadr_updated_report.oadrCancelReport
        if oadr_cancel_report:
            self.cancel_report(
                oadr_cancel_report.reportRequestID,
                acknowledge=False,
                request_id=request_id,
            )

    def handle_oadr_cancel_report(self, request_id, oadr_cancel_report):
        """
        The VTN responded to an oadrPoll by requesting a report cancellation.

        Respond by canceling the report, then send oadrCanceledReport to the VTN.

        :param oadr_cancel_report: the oadrCancelReportType element
        """
        self.cancel_report(
            oadr_cancel_report.reportRequestID, acknowledge=True, request_id=request_id
        )

    def handle_oadr_response(self, request_id, oadr_response):
        """
        The VTN has acknowledged a VEN request such as oadrCreatedReport.

        No response is needed.

        :param oadr_response: The VTN's request.
        """
        pass

    def create_or_update_reports(self, report_list, request_id):
        """
        Process report creation/update requests from the VTN (which could have arrived in different payloads).

        The requests could have arrived in response to a poll,
        or they could have been part of an oadrRegisteredReport response.

        Create/Update reports, and publish info about them on the volttron message bus.
        Send an oadrCreatedReport response to the VTN for each report.

        :param report_list: A iterable of oadrReportRequest. Can be None.
        """

        def get_default_interval(report_specifier, report_params):
            if "report_interval_secs_default" in report_params:
                try:
                    return int(report_params["report_interval_secs_default"])
                except ValueError:
                    raise ValueError(
                        f"Default report interval {report_params['report_interval_secs_default']}"
                        f" for report {report_specifier} is not an integer number of seconds"
                    )
            else:
                return None

        @orm.db_session
        def _create_pony_report(report_request) -> database.Report:
            """ Create a report in the databse."""
            specifier_id = extractors.report_extract_specifier_id(report_request)
            if specifier_id not in self.report_parameters:
                raise exceptions.BadDataError(
                    f"No parameters found for report with specifier ID {specifier_id}"
                )
            report_params = self.report_parameters[specifier_id]

            report = extractors.decode_ei_report(
                eiReport=report_request,
                default_interval=get_default_interval(specifier_id, report_params),
            )

            pony_report = database.Report(
                name=report_params.get("report_name"),
                telemetry_parameters=json.dumps(
                    report_params.get("telemetry_parameters")
                ),
                start_time=report.start_time,
                end_time=report.end_time,
                duration=report.duration,
                interval_secs=report.interval_secs,
                granularity_secs=report.granularity_secs,
                request_id=report.request_id,
                specifier_id=report.specifier_id,
            )
            return pony_report

        @orm.db_session
        def _update_pony_report(new_report_data, existing_report_id):
            """If the report changed, update its parameters in the database"""
            report = database.Report.get(id=existing_report_id)
            report.request_id = new_report_data.request_id
            report.specifier_id = new_report_data.specifier_id
            report.start_time = new_report_data.start_time
            report.end_time = new_report_data.end_time
            report.interval_secs = new_report_data.interval_secs
            return report

        @orm.db_session
        def cancel_rpt(rpt):
            """A report cancellation was received. Process it and notify interested parties."""
            rpt.status = enums.ReportStatus.CANCELED.value

        oadr_report_request_ids = []
        _LOGGER.debug(
            "Processing report creating/update requests from VTN with request id {}".format(
                request_id
            )
        )
        _LOGGER.debug(
            "Processing report creating/update requests from VTN with report list {}".format(
                report_list
            )
        )
        try:
            if report_list:
                for oadr_report_request in report_list:
                    report_data = extractors.decode_ei_report(oadr_report_request)
                    existing_report = self.get_report_for_report_specifier_id(
                        report_data.specifier_id
                    )
                    if existing_report:
                        # if the report already exists, update it
                        _update_pony_report(report_data, existing_report.id)
                    else:
                        _create_pony_report(oadr_report_request)

        #             if temp_report.status == enums.ReportStatus.CANCELED.value:
        #                 if existing_report:
        #                     oadr_report_request_ids.append(temp_report.request_id)
        #                     cancel_rpt(existing_report)
        #                     self.send_oadr_created_report(oadr_report_request)
        #                     _LOGGER.debug(
        #                         "Cancelled an existing report as VTN sent CANCELLED: {}".format(
        #                             existing_report
        #
        #          ),
        #                         exc_info=True,
        #                     )
        #                 else:
        #                     # Received notification of a new report, but it's already canceled. Take no action.
        #                     _LOGGER.debug(
        #                         "Received notification of a new report, but it's already canceled. Take no action.: {}".format(
        #                             existing_report
        #                         ),
        #                         exc_info=True,
        #                     )
        #             else:
        #                 oadr_report_request_ids.append(temp_report.request_id)

        #                 if temp_report.report_specifier_id == "METADATA":
        #                     # Rule 301/327: If the request's specifierID is 'METADATA', send an oadrRegisterReport.
        #                     self.send_oadr_created_report(oadr_report_request)
        #                     self.send_oadr_register_report()
        #                     _LOGGER.debug(
        #                         "Report was METADATA report, acknowledging registration to VTN: ".format(
        #                             temp_report
        #                         ),
        #                         exc_info=True,
        #                     )
        #                 elif existing_report:
        #                     update_rpt(temp_report, existing_report)
        #                     self.send_oadr_created_report(oadr_report_request)
        #                     _LOGGER.debug(
        #                         "Report already exists, updating with : ".format(
        #                             temp_report
        #                         ),
        #                         exc_info=True,
        #                     )
        #                 else:
        #                     create_rpt(temp_report)
        #                     self.send_oadr_created_report(oadr_report_request)
        #                     _LOGGER.debug(
        #                         "Report does not exist, creating : ".format(
        #                             temp_report
        #                         ),
        #                         exc_info=True,
        #                     )
        except exceptions.OpenADRInterfaceException as err:
            # If a VTN message contains a mix of valid and invalid reports, respond to the valid ones.
            # Don't reject the entire message due to an invalid report.
            _LOGGER.warning("Report error: {}".format(err), exc_info=True)
            self.send_oadr_response(
                err.message, err.error_code or response_codes.OADR_BAD_DATA, request_id
            )
        except Exception as err:
            _LOGGER.warning(
                "Unanticipated error during report processing: {}".format(err),
                exc_info=True,
            )
            self.send_oadr_response(str(err), response_codes.OADR_BAD_DATA, request_id)

        all_active_reports = self._active_reports
        for agent_report in all_active_reports:
            if agent_report.request_id not in oadr_report_request_ids:
                # If the VTN's request omitted an active report, treat it as an implied cancellation.
                report_request_id = agent_report.request_id
                _LOGGER.debug(
                    "Report request ID {} not sent by VTN, canceling the report.".format(
                        report_request_id
                    )
                )
                self.cancel_report(
                    report_request_id, acknowledge=True, request_id=request_id
                )

    @orm.db_session
    def cancel_report(
        self, report_request_id: str, request_id: str, acknowledge: bool = False
    ):
        """
        The VTN asked to cancel a report, in response to either report telemetry or an oadrPoll. Cancel it.

        :param report_request_id: The report_request_id of the report to be canceled.
        :param acknowledge: If True, send an oadrCanceledReport acknowledgment to the VTN.
        """
        if report_request_id is None:
            raise exceptions.OpenADRInterfaceException(
                "Missing oadrCancelReport.reportRequestID", response_codes.OADR_BAD_DATA
            )
        report = self.get_report_for_report_request_id(report_request_id)
        try:
            report.status = enums.ReportStatus.CANCELED.value
            if acknowledge:
                self.send_oadr_canceled_report(report_request_id, request_id)
        except:
            _LOGGER.warning(
                "The VEN got asked to cancel a report that it doesn't have. Did nothing."
            )

    # ***************** Send Requests from the VEN to the VTN ********************

    def send_oadr_poll(self):
        """Send oadrPoll to the VTN."""
        _LOGGER.debug("VEN: oadrPoll")
        # OADR rule 37: The VEN must support the PULL implementation.
        self._last_poll = get_aware_utc_now()
        builder = builders.OadrPollBuilder(ven_id=self.ven_id)
        self.send_vtn_request(POLL, "oadrPoll", builder.build())

    def send_oadr_query_registration(self):
        """Send oadrQueryRegistration to the VTN."""
        _LOGGER.debug("VEN: oadrQueryRegistration")
        builder = builders.OadrQueryRegistrationBuilder()
        self.send_vtn_request(EIREGISTERPARTY, "oadrQueryRegistration", builder.build())

    def send_oadr_create_party_registration(self, ven_name):
        """Send oadrCreatePartyRegistration to the VTN."""
        _LOGGER.debug("VEN: oadrCreatePartyRegistration")
        send_signature = self.security_level == "high"
        # OADR rule 404: If the VEN hasn't registered before, venID and registrationID should be empty.
        builder = builders.OadrCreatePartyRegistrationBuilder(
            ven_id=None, xml_signature=send_signature, ven_name=ven_name
        )
        self.send_vtn_request(
            EIREGISTERPARTY, "oadrCreatePartyRegistration", builder.build()
        )

    def send_oadr_request_event(self):
        """Send oadrRequestEvent to the VTN."""
        _LOGGER.debug("VEN: oadrRequestEvent")
        builder = builders.OadrRequestEventBuilder(ven_id=self.ven_id)
        self.send_vtn_request(EIEVENT, "oadrRequestEvent", builder.build())

    def send_oadr_created_event(
        self, event: database.Event, error_code: str = None, error_message: str = None
    ):
        """
        Send oadrCreatedEvent to the VTN.

        :param event: The event that is the subject of the request.
        :param error_code: eventResponse error code. Used when reporting event protocol errors.
        :param error_message: eventResponse error message. Used when reporting event protocol errors.
        """
        _LOGGER.debug("VEN: oadrCreatedEvent")
        builder = builders.OadrCreatedEventBuilder(
            event=event,
            ven_id=self.ven_id,
            error_code=error_code,
            error_message=error_message,
        )
        self.send_vtn_request(EIEVENT, "oadrCreatedEvent", builder.build())

    def send_oadr_register_report(self):
        """
        Send oadrRegisterReport (METADATA) to the VTN.

        Sample oadrRegisterReport from the OpenADR Program Guide:

            <oadr:oadrRegisterReport ei:schemaVersion="2.0b">
                <pyld:requestID>RegReq120615_122508_975</pyld:requestID>
                <oadr:oadrReport>
                    --- See oadr_report() ---
                </oadr:oadrReport>
                <ei:venID>ec27de207837e1048fd3</ei:venID>
            </oadr:oadrRegisterReport>
        """
        _LOGGER.debug("VEN: oadrRegisterReport")

        builder = builders.OadrRegisterReportBuilder(
            reports=self.metadata_reports(), ven_id=self.ven_id
        )
        self.send_vtn_request(EIREPORT, "oadrRegisterReport", builder.build())

    @orm.db_session
    def send_oadr_update_report(self, report_id: int):
        """
        Send telemetry for a given report to the VTN using oadrUpdateReport.

        Sample oadrUpdateReport from the OpenADR Program Guide:

            <oadr:oadrUpdateReport ei:schemaVersion="2.0b">
                <pyld:requestID>ReportUpdReqID130615_192730_445</pyld:requestID>
                <oadr:oadrReport>
                    --- See OadrUpdateReportBuilder ---
                </oadr:oadrReport>
                <ei:venID>VEN130615_192312_582</ei:venID>
            </oadr:oadrUpdateReport>
        """
        report = database.Report.get(id=report_id)
        _LOGGER.debug("VEN: oadrUpdateReport (report {})".format(report.request_id))
        self.oadr_current_service = EIREPORT
        telemetry = self.get_new_telemetry_for_report(report.specifier_id)
        _LOGGER.debug("VEN: sending telemetry {}".format(telemetry))
        builder = builders.OadrUpdateReportBuilder(
            report=report,
            telemetry=telemetry,
            online=self.ven_online,
            manual_override=self.ven_manual_override,
            ven_id=self.ven_id,
        )
        self.send_vtn_request(EIREPORT, "oadrUpdateReport", builder.build())
        report.last_report = datetime.utcnow()

    def send_oadr_created_report(self, report_request):
        """
            Send oadrCreatedReport to the VTN.

        @param report_request: (oadrReportRequestType) The VTN's report request.
        """
        _LOGGER.debug("VEN: oadrCreatedReport")
        builder = builders.OadrCreatedReportBuilder(
            report_request_id=report_request.reportRequestID,
            ven_id=self.ven_id,
            pending_report_request_ids=self.get_pending_report_request_ids(),
        )
        self.send_vtn_request(EIREPORT, "oadrCreatedReport", builder.build())

    def send_oadr_canceled_report(self, report_request_id, request_id):
        """
            Send oadrCanceledReport to the VTN.

        @param report_request_id: (string) The reportRequestID of the report that has been canceled.
        """
        _LOGGER.debug("VEN: oadrCanceledReport")
        builder = builders.OadrCanceledReportBuilder(
            request_id=request_id,
            report_request_id=report_request_id,
            ven_id=self.ven_id,
            pending_report_request_ids=self.get_pending_report_request_ids(),
        )
        self.send_vtn_request(EIREPORT, "oadrCanceledReport", builder.build())

    def send_oadr_response(
        self, description: str, code: str, request_id: Optional[str]
    ):
        """
        Send an oadrResponse to the VTN.
        """
        _LOGGER.debug("VEN: oadrResponse")
        builder = builders.OadrResponseBuilder(
            response_code=code,
            response_description=description,
            request_id=request_id or "0",
            ven_id=self.ven_id,
        )
        self.send_vtn_request(POLL, "oadrResponse", builder.build())

    def send_vtn_request(self, service, request_name, request_object):
        """
            Send a request to the VTN. If the VTN returns a non-empty response, service that request.

            Wrap the request in a SignedObject and then in Payload XML, and post it to the VTN via HTTP.
            If using high security, calculate a digital signature and include it in the request payload.

        @param request_name: (string) The name of the SignedObject attribute where the request is attached.
        @param request_object: (various oadr object types) The request to send.
        """
        signed_object = oadr_20b.oadrSignedObject(**{request_name: request_object})
        try:
            # Export the SignedObject as an XML string.
            buff = StringIO()
            signed_object.export(buff, 1, pretty_print=True)
            signed_object_serialised_xml = buff.getvalue()
        except Exception as err:
            raise exceptions.OpenADRInterfaceException(
                "Error exporting the SignedObject: {}".format(err), None
            )

        if self.security_level == "high":
            try:
                signature_lxml, signed_object_lxml = self.calculate_signature(
                    signed_object_serialised_xml
                )
            except Exception as err:
                raise exceptions.OpenADRInterfaceException(
                    "Error signing the SignedObject: {}".format(err), None
                )
            payload_lxml = self.payload_element(signature_lxml, signed_object_lxml)
            try:
                # Verify that the payload, with signature, is well-formed and can be validated.
                signxml.XMLVerifier().verify(payload_lxml, ca_pem_file=self.vtn_ca_cert)
            except Exception as err:
                raise exceptions.OpenADRInterfaceException(
                    "Error verifying the SignedObject: {}".format(err), None
                )
        else:
            signed_object_lxml = etree_.fromstring(signed_object_serialised_xml)
            payload_lxml = self.payload_element(None, signed_object_lxml)

        if self.log_xml:
            from bs4 import BeautifulSoup

            xml = etree_.tostring(payload_lxml, pretty_print=True)
            bs = BeautifulSoup(xml, "xml")
            _LOGGER.debug("VEN PAYLOAD:")
            _LOGGER.debug(f"\n{bs.prettify()}")

        # Post payload XML to the VTN as an HTTP request. Return the VTN's response, if any.
        endpoint = self.vtn_address + service
        payload_xml = etree_.tostring(payload_lxml)
        # OADR rule 53: If simple HTTP mode is used, send the following headers: Host, Content-Length, Content-Type.
        # The EPRI VTN server responds with a 400 "bad request" if a "Host" header is sent.

        try:
            payload_xml = etree_.tostring(payload_lxml)
            # OADR rule 53: If simple HTTP mode is used, send the following headers: Host, Content-Length, Content-Type.
            # The EPRI VTN server responds with a 400 "bad request" if a "Host" header is sent.
            _LOGGER.debug("Posting VEN request to {}".format(endpoint))
            response = requests.post(
                endpoint,
                cert=self.client_pem_bundle,
                data=payload_xml,
                headers={
                    # "Host": endpoint,
                    "Content-Length": str(len(payload_xml)),
                    "Content-Type": "application/xml",
                },
            )
            http_code = response.status_code
            if http_code == 200:
                if len(response.content) > 0:
                    self.service_vtn_request(response.content)
                else:
                    _LOGGER.warning("Received zero-length request from VTN")
            elif http_code == 204:
                # Empty response received. Take no action.
                _LOGGER.debug("Empty response received from {}".format(endpoint))
            else:
                _LOGGER.error(
                    "Error in http request to {}: response={}".format(
                        endpoint, http_code
                    ),
                    exc_info=True,
                )
                raise exceptions.OpenADRInterfaceException(
                    "Error in VTN request: {}".format(http_code)
                    + ":"
                    + str(response.content),
                    None,
                )
        except ConnectionError:
            _LOGGER.warning(
                "ConnectionError in http request to {} (is the VTN offline?)".format(
                    endpoint
                )
            )
            return None
        except Exception as err:
            raise exceptions.OpenADRInterfaceException(
                "Error posting OADR XML: {}".format(err), None
            )

    # ***************** Event database Requests ********************
    @property
    @orm.db_session
    def _unresponded_events(self) -> List[database.Event]:
        return orm.select(
            e for e in database.Event if e.status == enums.EventStatus.UNRESPONDED.value
        )[:]

    @property
    @orm.db_session
    def _near_events(self) -> List[database.Event]:
        return orm.select(
            e for e in database.Event if e.status == enums.EventStatus.NEAR.value
        )[:]

    @property
    @orm.db_session
    def _far_events(self) -> List[database.Event]:
        return orm.select(
            e for e in database.Event if e.status == enums.EventStatus.FAR.value
        )[:]

    @property
    @orm.db_session
    def _active_events(self) -> List[database.Event]:
        """Return a list of events that are currently in progress
           status is active.
        """
        query = orm.select(
            e for e in database.Event if e.status == enums.EventStatus.ACTIVE.value
        )
        query = query.filter(
            lambda event: event.opt_type != enums.OptType.OPT_OUT.value
        )
        return query[:]

    @property
    def _active_or_pending_events(self) -> List[database.Event]:
        """ Returns active or pending events, in the sequence
        active, near, far, unresponded"""
        events: List[database.Event] = []
        if self._active_events:
            events += self._active_events
        if self._near_events:
            events += self._near_events
        if self._far_events:
            events += self._far_events
        if self._unresponded_events:
            events += self._unresponded_events
        return events

    @property
    def active_or_pending_events(self) -> List[dict]:
        events_dict_list = []
        for event in self._active_or_pending_events:
            event_dict = event.to_dict()
            events_dict_list.append(event_dict)
        return events_dict_list

    @property
    def active_events(self) -> List[dict]:
        events_dict_list = []
        for event in self._active_events:
            event_dict = event.to_dict()
            events_dict_list.append(event_dict)
        return events_dict_list

    @property
    def is_event_in_progress(self) -> bool:
        """ Is an event in progress? """
        return len(self._active_events) > 0

    @orm.db_session
    def get_event_for_id(self, event_id: str) -> Optional[database.Event]:
        """ Query the DB for the event with the given ID """
        return database.Event.get(event_id=event_id)

    @orm.db_session
    def _set_event_status(self, event_id, status_enum: enums.EventStatus):

        """
        Transition an event from its existing status to a new one.

        :raises exceptions.InvalidStatusException:
        """
        event = self.get_event_for_id(event_id)
        if not isinstance(status_enum, enums.EventStatus):
            raise exceptions.InvalidStatusException(
                "status provided is not a valid EventStatus"
            )
        if status_enum == enums.EventStatus.ACTIVE:
            if event.opt_type != enums.OptType.OPT_IN.value:
                raise exceptions.InvalidStatusException(
                    "can't transition status to ACTIVE if event not opted in"
                )
        _LOGGER.debug(
            f"Transitioning status to {status_enum} for event ID {event.event_id}"
        )
        event.status = status_enum.value

    @property
    @orm.db_session
    def _active_reports(self) -> List[database.Report]:
        query = orm.select(
            r for r in database.Report if r.status == enums.ReportStatus.ACTIVE.value
        )
        return query[:]

    @property
    @orm.db_session
    def _inactive_reports(self) -> List[database.Report]:
        query = orm.select(
            r for r in database.Report if r.status == enums.ReportStatus.INACTIVE.value
        )
        return query[:]

    @property
    def active_or_pending_reports(self) -> List[database.Report]:
        """Return a list of reports that are neither COMPLETED nor CANCELED."""
        reports: List[database.Report] = []
        if self._active_reports:
            reports += self._active_reports
        if self._inactive_reports:
            reports += self._inactive_reports

        return reports

    @orm.db_session
    def set_report_status(self, report: database.Report, status: enums.ReportStatus):
        if report.status != status.value:
            _LOGGER.debug(
                f"Transitioning status to {status} for report request ID {report.request_id}"
            )
            report.status = status.value

    @orm.db_session
    def get_report_for_report_request_id(
        self, report_request_id: str
    ) -> Optional[database.Report]:
        """Return the Report with request ID report_request_id, or None if not found."""
        return database.Report.get(request_id=report_request_id)

    @orm.db_session
    def get_report_for_report_specifier_id(
        self, specifier_id: str
    ) -> Optional[database.Report]:
        """Return the Report with specifier ID specifier_id, or None if not found."""
        return database.Report.get(specifier_id=specifier_id)

    def get_pending_report_request_ids(self):
        """Return a list of reportRequestIDs for each active report."""
        # OpenADR rule 329: Include all current report request IDs in the oadrPendingReports list.
        return [r.report_request_id for r in self._active_reports]

    def metadata_reports(self):
        """Return an Report instance containing telemetry metadata for each report definition in agent config."""
        return [
            self.metadata_report(rpt_name) for rpt_name in self.report_parameters.keys()
        ]

    @orm.db_session
    def metadata_report(self, specifier_id: str) -> database.Report:
        """Return an Report instance for the indicated specifier_id, or None if its' not in agent config."""
        params = self.report_parameters.get(specifier_id, None)

        if self.get_report_for_report_specifier_id(specifier_id):
            """ If there is already a report with this specifier id, use it! """
            report = self.get_report_for_report_specifier_id(specifier_id)

        else:
            report = database.Report(request_id="", specifier_id=specifier_id)
            report.name = params.get("report_name_metadata", None)
            report.telemetry_parameters = json.dumps(
                params.get("telemetry_parameters", None)
            )
            report.report_specifier_id = specifier_id
            report.status = enums.ReportStatus.INACTIVE.value

        # Report default interval seconds is not stored for some reason so need to re-set it here on existing reports!
        # Is a default interval required by spec?
        try:
            interval_secs = int(params.get("report_interval_secs_default", None))
        except ValueError:
            error_msg = "Default report interval {} is not an integer number of seconds".format(
                params.get("report_interval_secs_default")
            )
            raise exceptions.OpenADRInternalException(error_msg)
        report.interval_secs = interval_secs

        return report

    @orm.db_session
    def get_new_telemetry_for_report(
        self, report_specifier_id: str
    ) -> List[database.TelemetryValues]:
        """Query for relevant telemetry that's arrived since the report was last sent to the VTN."""
        report = self.get_report_for_report_specifier_id(report_specifier_id)
        telemetry = orm.select(
            t
            for t in database.TelemetryValues
            if t.report_specifier_id == report.specifier_id
        )
        if report.last_report is None:
            return list(telemetry)

        telemetry = telemetry.filter(lambda t: t.created > report.last_report)
        return list(telemetry)

    @orm.db_session
    def add_telemetry_json(
        self, report_specifier_id: str, values: Dict, start_time=None, end_time=None
    ):

        """New telemetry (JSON) has been received. Add it to the database."""
        if self.report_parameters[report_specifier_id] is None:
            raise exceptions.BadDataError(
                f"No parameters found for report with specifier ID {report_specifier_id}\n"
                + "Registered reports are: "
                + str(self.report_parameters.keys())
            )
        telemetry = database.TelemetryValues(
            report_specifier_id=report_specifier_id,
            start_time=start_time,
            end_time=end_time,
            values=json.dumps(values),
        )
        return telemetry

    # ***************** Utility Methods ********************

    @staticmethod
    def payload_element(
        signature_lxml: Optional[etree_.Element], signed_object_lxml: etree_.Element
    ) -> etree_.Element:
        """
            Construct and return an XML element for Payload.

            Append a child Signature element if one is provided.
            Append a child SignedObject element.

        :param signature_lxml: Signature element.
        :param signed_object_lxml: SignedObject element.
        :return: Payload element.
        """
        payload = etree_.Element(
            "{http://openadr.org/oadr-2.0b/2012/07}oadrPayload",
            nsmap=signed_object_lxml.nsmap,
        )
        if signature_lxml:
            payload.append(signature_lxml)
        payload.append(signed_object_lxml)
        return payload

    @staticmethod
    def calculate_signature(
        self, serialised_xml: str
    ) -> Tuple[etree_.Element, etree_.Element]:
        """
        Calculate a digital signature for the SignedObject to be sent to the VTN.

        Returns a tuple of
            1. An object representing the root of the XML tree containing the signature and the payload data.
            2. The XML tree parsed from serialised_xml
        """

        private_key, certificate = pem.parse_file(self.client_pem_bundle)
        xml_tree = etree_.fromstring(serialised_xml)
        xml_tree.set("Id", "signedObject")
        # Use XMLSigner to create a Signature.
        # Use "detached method": the signature lives alonside the signed object in the XML element tree.
        # Use c14n "exclusive canonicalization": the signature is independent of namespace inclusion/exclusion.
        signer = signxml.XMLSigner(
            method=signxml.methods.detached,
            c14n_algorithm="http://www.w3.org/2001/10/xml-exc-c14n#",
        )
        signature_root = signer.sign(
            xml_tree,
            key=private_key.as_bytes(),
            cert=certificate.as_bytes(),
            key_name="123",
        )
        # This generated Signature lacks the ReplayProtect property described in
        # OpenADR profile spec section 10.6.3.
        return signature_root, xml_tree

    def json_object(self, obj):
        """Ensure that an object is valid JSON by dumping it with json_converter and then reloading it."""
        obj_string = json.dumps(obj, default=self.json_converter)
        obj_json = json.loads(obj_string)
        return obj_json

    @staticmethod
    def json_converter(object_to_dump):
        """When calling json.dumps, convert datetime instances to strings."""
        if isinstance(object_to_dump, datetime):
            return object_to_dump.__str__()



from . import response_codes


class OpenADRInterfaceException(Exception):
    """
    Use this exception when an error should be sent to the VTN as an oadrResponse
    payload.

    :param message:    message to send back to the VTN (and potentially log)
    :param error_code: error code taken from ``pyoadr_ven.response_codes``
    """

    def __init__(self, message, error_code, *args):
        super().__init__(message, *args)
        self.error_code = error_code
        self.message = message


class BadDataError(OpenADRInterfaceException):
    def __init__(self, message, *args):
        super().__init__(message, response_codes.OADR_BAD_DATA, *args)


class OpenADRInternalException(Exception):
    """
    Use this exception when an error should be logged but not sent to the VTN.
    """

    def __init__(self, message, *args):
        super().__init__(message, *args)
        self.message = message


class InvalidStatusException(OpenADRInternalException):
    def __init__(self, message, *args):
        super().__init__(message, *args)



from enum import Enum


class ExtendedEnum(Enum):
    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))


class EventStatus(ExtendedEnum):
    UNRESPONDED = "unresponded"
    FAR = "far"
    NEAR = "near"
    ACTIVE = "active"
    COMPLETED = "completed"
    CANCELED = "cancelled"


class ReportStatus(ExtendedEnum):
    INACTIVE = "inactive"
    ACTIVE = "active"
    COMPLETED = "completed"
    CANCELED = "cancelled"


class OptType(ExtendedEnum):
    OPT_IN = "optIn"
    OPT_OUT = "optOut"
    NONE = "none"



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2017, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}
import json
import logging
import uuid
from datetime import timedelta

import isodate

from . import oadr_20b
from .response_codes import OADR_VALID_RESPONSE
from .utils import get_aware_utc_now

_LOGGER = logging.getLogger(__name__)
SCHEMA_VERSION = "2.0b"


class OadrBuilder(object):
    """Abstract superclass. Build oadr model objects to send to the VTN."""

    def __init__(self, request_id=None, ven_id=None, **kwargs):
        self.request_id = request_id
        self.ven_id = ven_id

    @staticmethod
    def create_request_id():
        return uuid.uuid1()


class OadrPollBuilder(OadrBuilder):
    def __init__(self, **kwargs):
        super(OadrPollBuilder, self).__init__(**kwargs)

    def build(self):
        return oadr_20b.oadrPollType(schemaVersion=SCHEMA_VERSION, venID=self.ven_id)


class OadrQueryRegistrationBuilder(OadrBuilder):
    def __init__(self, **kwargs):
        super(OadrQueryRegistrationBuilder, self).__init__(**kwargs)

    def build(self):
        return oadr_20b.oadrQueryRegistrationType(
            schemaVersion=SCHEMA_VERSION, requestID=self.create_request_id()
        )


class OadrCreatePartyRegistrationBuilder(OadrBuilder):
    def __init__(self, xml_signature=None, ven_name=None, **kwargs):
        super(OadrCreatePartyRegistrationBuilder, self).__init__(**kwargs)
        self.xml_signature = xml_signature
        self.ven_name = ven_name

    def build(self):
        return oadr_20b.oadrCreatePartyRegistrationType(
            schemaVersion=SCHEMA_VERSION,
            requestID=self.create_request_id(),
            registrationID=None,
            venID=self.ven_id,
            oadrProfileName="2.0b",
            oadrTransportName="simpleHttp",
            oadrTransportAddress=None,
            oadrReportOnly=False,
            oadrXmlSignature=self.xml_signature,
            oadrVenName=self.ven_name,
            oadrHttpPullModel=True,
        )


class OadrRequestEventBuilder(OadrBuilder):
    def __init__(self, **kwargs):
        super(OadrRequestEventBuilder, self).__init__(**kwargs)

    def build(self):
        ei_request_event = oadr_20b.eiRequestEvent(
            requestID=self.create_request_id(), venID=self.ven_id
        )
        return oadr_20b.oadrRequestEventType(
            schemaVersion=SCHEMA_VERSION, eiRequestEvent=ei_request_event
        )


class OadrCreatedEventBuilder(OadrBuilder):

    """
        Construct an oadrCreatedEvent to return in response to an oadrDistributeEvent or oadrCreateEvent.

        If an error occurs while processing an events in the VTN's payload, return an oadrCreatedEvent
        response in which the eiResponse has a normal/valid (200) code,
        and the eventResponse has a responseCode containing the error code.
    """

    def __init__(self, event=None, error_code=None, error_message=None, **kwargs):
        super(OadrCreatedEventBuilder, self).__init__(**kwargs)
        self.event = event
        self.error_code = error_code
        self.error_message = error_message

    def build(self):
        ei_response = oadr_20b.EiResponseType(
            responseCode=OADR_VALID_RESPONSE, requestID=None
        )
        qualified_event_id = oadr_20b.QualifiedEventIDType(
            eventID=self.event.event_id,
            modificationNumber=self.event.modification_number,
        )
        # OADR rule 42: the requestID from the oadrDistributeEvent must be re-used by the eventResponse.
        event_response = oadr_20b.eventResponseType(
            responseCode=self.error_code or OADR_VALID_RESPONSE,
            responseDescription=self.error_message,
            requestID=self.event.request_id,
            qualifiedEventID=qualified_event_id,
            optType=self.event.opt_type,
        )
        # OADR rule 25, 35: eventResponses is required except when eiResponse indicates failure.
        event_responses = oadr_20b.eventResponses()
        event_responses.add_eventResponse(event_response)
        ei_created_event = oadr_20b.eiCreatedEvent(
            eiResponse=ei_response, eventResponses=event_responses, venID=self.ven_id
        )
        return oadr_20b.oadrCreatedEventType(eiCreatedEvent=ei_created_event)


class OadrReportBuilder(OadrBuilder):
    def __init__(
        self,
        report=None,
        report_request_id=None,
        pending_report_request_ids=None,
        **kwargs,
    ):
        super(OadrReportBuilder, self).__init__(**kwargs)
        self.report = report
        self.request_id = report_request_id
        self.pending_report_request_ids = pending_report_request_ids


class OadrRegisterReportBuilder(OadrReportBuilder):
    def __init__(self, reports=None, **kwargs):
        super(OadrRegisterReportBuilder, self).__init__(**kwargs)
        self.reports = reports

    def build(self):
        oadr_reports = [
            self.build_metadata_oadr_report(report) for report in self.reports
        ]
        return oadr_20b.oadrRegisterReportType(
            schemaVersion=SCHEMA_VERSION,
            requestID=self.create_request_id(),
            oadrReport=oadr_reports,
            venID=self.ven_id,
            reportRequestID=None,
        )

    def build_metadata_oadr_report(self, report):
        descriptions = []
        for tel_vals in json.loads(report.telemetry_parameters).values():
            # Rule 305: For TELEMETRY_USAGE reports, units in reportDescription.itemBase should be powerReal.
            if tel_vals["units"] == "powerReal":
                item_base = oadr_20b.PowerRealType(
                    itemDescription="RealPower",
                    itemUnits="W",
                    siScaleCode=None,
                    powerAttributes=None,
                )
            else:
                item_base = None
            min_freq, max_freq = tel_vals["min_frequency"], tel_vals["max_frequency"]
            desc = oadr_20b.oadrReportDescriptionType(
                rID=tel_vals["r_id"],
                reportType=tel_vals["report_type"],
                readingType=tel_vals["reading_type"],
                itemBase=item_base,
                oadrSamplingRate=self.build_sampling_rate(min_freq, max_freq),
            )
            descriptions.append(desc)
        rpt_interval_duration = isodate.duration_isoformat(
            timedelta(seconds=report.interval_secs)
        )
        return oadr_20b.oadrReportType(
            duration=oadr_20b.DurationPropType(rpt_interval_duration),
            oadrReportDescription=descriptions,
            reportRequestID=None,
            reportSpecifierID=report.specifier_id,
            reportName=report.name,
            createdDateTime=get_aware_utc_now(),
        )

    @staticmethod
    def build_sampling_rate(min_freq, max_freq):
        min_sampling_rate = isodate.duration_isoformat(timedelta(seconds=min_freq))
        max_sampling_rate = isodate.duration_isoformat(timedelta(seconds=max_freq))
        sampling_rate = oadr_20b.oadrSamplingRateType(
            oadrMinPeriod=min_sampling_rate,
            oadrMaxPeriod=max_sampling_rate,
            oadrOnChange=False,
        )
        return sampling_rate


class OadrUpdateReportBuilder(OadrReportBuilder):
    def __init__(self, telemetry=None, online=None, manual_override=None, **kwargs):
        super(OadrUpdateReportBuilder, self).__init__(**kwargs)
        self.telemetry = telemetry
        self.online = online
        self.manual_override = manual_override

    def build(self):
        """
            Return an oadrReport containing telemetry updates.

            A typical XML element structure is:

            <oadr:oadrReport>
                <xcal:dtstart>
                    <xcal:date-time>2017-12-06T21:33:32Z</xcal:date-time>
                </xcal:dtstart>
                <xcal:duration>
                    <oadr:duration>PT0S</oadr:duration>
                </xcal:duration>
                <strm:intervals>
                    <ei:interval>
                        <xcal:dtstart>
                            <xcal:date-time>2017-12-06T21:33:08.423684Z</xcal:date-time>
                        </xcal:dtstart>
                        <xcal:duration>
                            <oadr:duration>PT30S</oadr:duration>
                        </xcal:duration>
                        <oadr:oadrReportPayload>
                            <ei:rID>baseline_power</ei:rID>
                            <ei:payloadFloat>
                                <ei:value>6.2</ei:value>
                            </ei:payloadFloat>
                        </oadr:oadrReportPayload>
                        <oadr:oadrReportPayload>
                            <ei:rID>actual_power</ei:rID>
                            <ei:payloadFloat>
                                <ei:value>5.44668467252</ei:value>
                            </ei:payloadFloat>
                        </oadr:oadrReportPayload>
                        <oadr:oadrReportPayload>
                            <ei:rID>Status</ei:rID>
                            <oadr:oadrPayloadResourceStatus>
                                <oadr:oadrOnline>true</oadr:oadrOnline>
                                <oadr:oadrManualOverride>false</oadr:oadrManualOverride>
                            </oadr:oadrPayloadResourceStatus>
                        </oadr:oadrReportPayload>
                    </ei:interval>
                </strm:intervals>
                <ei:reportRequestID>RR_916fd571c0657070575a</ei:reportRequestID>
                <ei:reportSpecifierID>telemetry</ei:reportSpecifierID>
                <ei:reportName>TELEMETRY_USAGE</ei:reportName>
                <ei:createdDateTime>2017-12-06T21:33:47.869392Z</ei:createdDateTime>
            </oadr:oadrReport>

        @return: (oadrReportType) An oadrReport.
        """

        oadr_report = oadr_20b.oadrReportType(
            dtstart=oadr_20b.dtstart(date_time=self.report.start_time),
            duration=oadr_20b.DurationPropType(self.report.iso_duration),
            reportRequestID=self.report.request_id,
            reportSpecifierID=self.report.specifier_id,
            reportName=self.report.name,
            intervals=self.build_intervals(),
            createdDateTime=get_aware_utc_now(),
        )
        return oadr_20b.oadrUpdateReportType(
            schemaVersion=SCHEMA_VERSION,
            requestID=self.create_request_id(),
            oadrReport=[oadr_report],
            venID=self.ven_id,
        )

    def build_intervals(self):
        """Build an intervals object, holding a list of Intervals to report in this update."""
        intervals = oadr_20b.intervals()

        if self.report.specifier_id == "telemetry_status":
            # OADR rule 331, 510: Build an Interval, in a telemetry_status report, giving current VEN status.
            interval_start = isodate.parse_datetime(self.report.iso_last_report)
            interval_duration = isodate.duration_isoformat(
                timedelta(seconds=self.report.interval_secs)
            )
            interval = oadr_20b.IntervalType(
                dtstart=oadr_20b.dtstart(date_time=interval_start),
                duration=oadr_20b.DurationPropType(duration=interval_duration),
            )
            if self.online is not None or self.manual_override is not None:
                payload = oadr_20b.oadrReportPayloadType(rID="Status")
                payload_status = oadr_20b.oadrPayloadResourceStatusType(
                    oadrOnline=self.online, oadrManualOverride=self.manual_override
                )
                payload_status.original_tagname_ = "oadrPayloadResourceStatus"
                payload.set_payloadBase(payload_status)
                interval.set_streamPayloadBase([payload])
            else:
                interval.set_streamPayloadBase([])
            intervals.add_interval(interval)

        else:  # used to be if self.report.specifier_id == "telemetry":
            # Build Intervals that report metrics.
            for telemetry_values in self.telemetry:
                intervals.add_interval(self.build_report_interval(telemetry_values))
        return intervals

    def build_report_interval(self, telemetry_values):
        """Build an Interval for a report timeframe that includes telemetry values gathered during that time."""

        def interval_start_time():
            """ OpenADR Spec rule #314
            """
            if telemetry_values.start_time:
                return telemetry_values.start_time
            else:
                return self.report.start_time

        # duration_isoformat can yield fractional seconds, e.g. PT59.9S, resulting in a warning during XML creation.
        interval_duration = isodate.duration_isoformat(telemetry_values.duration)
        interval = oadr_20b.IntervalType(
            dtstart=oadr_20b.dtstart(date_time=interval_start_time()),
            duration=oadr_20b.DurationPropType(duration=interval_duration),
        )
        _LOGGER.info(f"telemetry values: {telemetry_values.to_dict()}")
        interval.set_streamPayloadBase(self.build_report_payload_list(telemetry_values))
        return interval

    def build_report_payload_list(self, telemetry_values):
        """Build a list of ReportPayloads containing current telemetry."""
        report_payload_list = []
        _LOGGER.warning((self.report.telemetry_parameters))
        try:
            for tel_val in json.loads(self.report.telemetry_parameters).values():
                payload = self.build_report_payload_float(
                    telemetry_values, tel_val["r_id"], tel_val["method_name"]
                )
                if payload:
                    report_payload_list.append(payload)
            _LOGGER.info(f"Report Payload List: {report_payload_list}")
            return report_payload_list
        except json.JSONDecodeError as err:
            _LOGGER.exception(
                f"Error json decoding the telemetry parameters - report {self.report.to_dict()}"
            )

    @staticmethod
    def build_report_payload_float(telemetry_values, r_id, method_name):
        """Build a single ReportPayload containing a PayloadFloat metric."""
        metric = telemetry_values.values_dict[method_name]
        if metric is None:
            payload = None
        else:
            payload = oadr_20b.oadrReportPayloadType(rID=r_id)
            payload.original_tagname_ = "oadrReportPayload"
            payload_float = oadr_20b.PayloadFloatType(value=metric)
            payload_float.original_tagname_ = "payloadFloat"
            payload.set_payloadBase(payload_float)
        return payload


class OadrCreatedReportBuilder(OadrReportBuilder):
    def __init__(self, **kwargs):
        super(OadrCreatedReportBuilder, self).__init__(**kwargs)

    def build(self):
        ei_response = oadr_20b.EiResponseType(
            responseCode=OADR_VALID_RESPONSE, requestID=self.create_request_id()
        )
        oadr_pending_reports = oadr_20b.oadrPendingReportsType(
            reportRequestID=self.pending_report_request_ids
        )
        return oadr_20b.oadrCreatedReportType(
            schemaVersion=SCHEMA_VERSION,
            eiResponse=ei_response,
            oadrPendingReports=oadr_pending_reports,
            venID=self.ven_id,
        )


class OadrCanceledReportBuilder(OadrReportBuilder):
    def __init__(self, **kwargs):
        super(OadrCanceledReportBuilder, self).__init__(**kwargs)

    def build(self):
        # Note that -- oddly -- this message does NOT contain the reportRequestID of the canceled report.
        # It does list all active reports in oadrPendingReports, though,
        # so the VTN has sufficient information to know which report was canceled.
        ei_response = oadr_20b.EiResponseType(
            responseCode=OADR_VALID_RESPONSE, requestID=self.request_id
        )
        oadr_pending_reports = oadr_20b.oadrPendingReportsType(
            reportRequestID=self.pending_report_request_ids
        )
        return oadr_20b.oadrCanceledReportType(
            schemaVersion=SCHEMA_VERSION,
            eiResponse=ei_response,
            oadrPendingReports=oadr_pending_reports,
            venID=self.ven_id,
        )


class OadrResponseBuilder(OadrBuilder):
    def __init__(self, response_code=None, response_description=None, **kwargs):
        super(OadrResponseBuilder, self).__init__(**kwargs)
        self.response_code = response_code
        self.response_description = response_description

    def build(self):
        ei_response = oadr_20b.EiResponseType(
            responseCode=self.response_code,
            responseDescription=self.response_description,
            requestID=self.request_id,
        )
        return oadr_20b.oadrResponseType(
            schemaVersion=SCHEMA_VERSION, eiResponse=ei_response, venID=self.ven_id
        )



import logging

import pytest
import responses

_LOGGER = logging.getLogger(__name__)

VTN_ADDRESS = "https://openadr-staging"
ENDPOINT_BASE = "/OpenADR2/Simple/2.0b/"
ENDPOINT = VTN_ADDRESS + ENDPOINT_BASE
EIEVENT = ENDPOINT + "EiEvent"
EIREPORT = ENDPOINT + "EiReport"
EIREGISTERPARTY = ENDPOINT + "EiRegisterParty"
POLL = ENDPOINT + "OadrPoll"

with open("tests/xml/ven_oadrRegisterReport.xml", "r") as f:
    oadrRegisterReport_body = f.read()

with open("tests/xml/vtn_oadrRegisterReport_response.xml", "r") as f:
    oadrRegisterReport_response_body = f.read()

pytestmark = pytest.mark.pony(db_session=False)


class TestRegisterReport:
    @responses.activate
    def test_register_report(self, mocked_agent):
        responses.add(
            responses.POST,
            EIREPORT,
            content_type="text/xml",
            body=oadrRegisterReport_body,
        )
        responses.add(
            responses.POST,
            EIREPORT,
            content_type="text/xml",
            body=oadrRegisterReport_response_body,
        )
        mocked_agent.send_oadr_register_report()
        assert responses.calls[0].request.url == EIREPORT



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2017, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}
import json
from datetime import datetime as dt
from typing import Optional
from typing import Type

from dateutil import parser

from .utils import format_timestamp
from .utils import get_aware_utc_now


class EiReport:
    """Model object for a report."""

    __tablename__ = "EiReport"

    __slots__ = (
        "iso_start_time",
        "iso_end_time",
        "duration",
        "iso_last_report",
        "name",
        "interval_secs",
        "granularity_secs",
        "telemetry_parameters",
        "created_on",
        "request_id",
        "report_request_id",
        "report_specifier_id",
        "status",
    )

    def __init__(
        self,
        request_id,
        report_request_id,
        report_specifier_id,
        granularity_secs=None,
        interval_secs=None,
        iso_start_time=None,
        iso_end_time=None,
        duration=None,
    ):
        self.iso_start_time: str = iso_start_time or ""  # ISO 8601 timestamp in UTC
        self.iso_end_time: str = iso_end_time or ""  # ISO 8601 timestamp in UTC
        self.duration: str = duration or ""  # ISO 8601 duration
        self.iso_last_report: str = ""  # ISO 8601 timestamp in UTC
        self.name: str = ""
        self.interval_secs: Optional[int] = interval_secs
        self.granularity_secs: Optional[int] = granularity_secs
        self.telemetry_parameters: str = ""
        self.created_on: Type(dt) = get_aware_utc_now()
        self.request_id: str = request_id
        self.report_request_id: str = report_request_id
        self.report_specifier_id: str = report_specifier_id
        self.status: str = "inactive"
        self.last_report: Type(dt) = get_aware_utc_now()

    def __str__(self):
        """Format the instance as a string suitable for trace display."""
        my_str = "{}: ".format(self.__class__.__name__)
        my_str += "report_request_id:{}; ".format(self.report_request_id)
        my_str += "report_specifier_id:{}; ".format(self.report_specifier_id)
        my_str += "start_time:{}; ".format(self.start_time)
        my_str += "end_time:{}; ".format(self.end_time)
        my_str += "status:{}; ".format(self.status)
        return my_str

    @property
    def start_time(self):
        return parser.parse(self.iso_start_time) if self.iso_start_time else None

    @property
    def end_time(self):
        return parser.parse(self.iso_end_time) if self.iso_end_time else None

    @property
    def last_report(self):
        return parser.parse(self.iso_last_report) if self.iso_last_report else None

    @start_time.setter
    def start_time(self, t):
        self.iso_start_time = format_timestamp(t) if t else None

    @end_time.setter
    def end_time(self, t):
        self.iso_end_time = format_timestamp(t) if t else None

    @last_report.setter
    def last_report(self, t):
        self.iso_last_report = format_timestamp(t) if t else None

    def is_active_or_pending(self):
        return self.status not in [self.STATUS_COMPLETED, self.STATUS_CANCELED]

    def as_json_compatible_object(self):
        """Format the object as JSON that will be returned in response to an RPC, or sent in a pub/sub."""
        return {attname: getattr(self, attname) for attname in self.__slots__}

    def copy_from_report(self, another_report):
        """(Selectively) Copy the contents of another_report to this one."""
        self.request_id = another_report.request_id
        self.report_request_id = another_report.report_request_id
        self.report_specifier_id = another_report.report_specifier_id
        self.start_time = another_report.start_time
        self.end_time = another_report.end_time
        self.duration = another_report.duration
        self.granularity_secs = another_report.granularity_secs
        # Do not copy created_on from another_report
        # Do not copy status from another_report
        # Do not copy last_report from another_report
        self.name = another_report.name
        self.interval_secs = another_report.interval_secs
        self.telemetry_parameters = another_report.telemetry_parameters


class EiTelemetryValues:
    """Model object for telemetry values."""

    __tablename__ = "EiTelemetryValues"

    __slots__ = (
        "created_on",
        "report_specifier_id",
        "iso_start_time",
        "iso_end_time",
        "values",
    )

    def __init__(
        self, report_specifier_id=None, start_time=None, end_time=None, values=None
    ):
        self.created_on: Type(dt) = get_aware_utc_now()
        self.report_specifier_id: str = report_specifier_id
        self.start_time: str = start_time
        self.end_time: str = end_time
        self.values: dict = values

    def __str__(self):
        """Format the instance as a string suitable for trace display."""
        my_str = "{}: ".format(self.__class__.__name__)
        my_str += "created_on:{}; ".format(self.created_on)
        my_str += "report_request_id:{}; ".format(self.report_request_id)
        my_str += "start_time:{} ".format(self.start_time)
        my_str += "end_time:{} ".format(self.end_time)
        my_str += "values:{} ".format(json.dumps(self.values))
        return my_str

    @property
    def start_time(self):
        return parser.parse(self.iso_start_time) if self.iso_start_time else None

    @property
    def end_time(self):
        return parser.parse(self.iso_end_time) if self.iso_end_time else None

    @start_time.setter
    def start_time(self, t):
        self.iso_start_time = format_timestamp(t) if t else None

    @end_time.setter
    def end_time(self, t):
        self.iso_end_time = format_timestamp(t) if t else None

    @classmethod
    def sample_values(cls):
        """Return a sample set of telemetry values for debugging purposes."""
        telemetry_values = cls()
        telemetry_values.report_specifier_id = "123"
        telemetry_values.values = {"asdasd": 123123}
        return telemetry_values

    def as_json_compatible_object(self):
        """Format the object as JSON that will be returned in response to an RPC, or sent in a pub/sub."""
        return {attname: getattr(self, attname) for attname in self.attribute_names}

    def get_duration(self):
        return self.end_time - self.start_time



import json
from datetime import datetime
from datetime import timedelta

import isodate
from pony import orm

from .enums import EventStatus
from .enums import OptType
from .enums import ReportStatus


db = orm.Database()


class Event(db.Entity):
    event_id = orm.Required(str)

    # This is the start time as specified by the VTN
    official_start = orm.Required(datetime)

    # If larger than 0, this is the largest amount of randomisation we can (must) apply
    # to official_start to get start_time (below).  This is stored in start_offset.
    #
    # If 0, no randomisation is applied and start_time == official_time.
    start_after = orm.Required(timedelta)

    # This is how long the event lasts
    duration = orm.Optional(timedelta)

    # This is our calculated random offset.
    # 0 ≤ start_offset ≤ start_after
    start_offset = orm.Required(timedelta)

    # start_time = official_start + start_offset (precalculated for efficiency)
    start_time = orm.Required(datetime)

    # end_time = start_time + duration (precalculated for efficiency)
    end_time = orm.Optional(datetime)

    request_id = orm.Optional(str, nullable=True)
    created = orm.Required(datetime, default=datetime.utcnow)
    signals = orm.Optional(str)
    status = orm.Required(str, default=EventStatus.UNRESPONDED.value)
    opt_type = orm.Optional(str, default=OptType.NONE.value)
    priority = orm.Optional(int)
    modification_number = orm.Required(int, default=0)
    test_event = orm.Required(bool, default=False)

    def __str__(self):
        """Format the instance as a string suitable for trace display."""
        event_string = f"{self.__class__.__name__}: "
        event_string += f"event_id:{self.event_id}; "
        event_string += f"start_time:{self.start_time}; "
        event_string += f"end_time:{self.end_time}; "
        event_string += f"request_id:{self.request_id}; "
        event_string += f"status:{self.status}; "
        event_string += f"opt_type: {self.opt_type};"
        event_string += f"priority:{self.priority}; "
        event_string += f"modification_number:{self.modification_number}; "
        event_string += f"signals:{self.signals}; "
        return event_string

    @property
    def is_active_or_pending(self):
        return self.status not in [
            EventStatus.COMPLETED.value,
            EventStatus.CANCELED.value,
        ]


class Report(db.Entity):
    request_id = orm.Optional(str)
    specifier_id = orm.Required(str)
    start_time = orm.Required(datetime, default=datetime.utcnow)
    end_time = orm.Optional(datetime)
    duration = orm.Optional(timedelta)
    name = orm.Optional(str, nullable=True)
    interval_secs = orm.Optional(int)
    granularity_secs = orm.Optional(int)
    telemetry_parameters = orm.Optional(str)
    created = orm.Required(datetime, default=datetime.utcnow)
    status = orm.Required(str, default=ReportStatus.INACTIVE.value)
    last_report = orm.Optional(datetime)

    @property
    def is_active_or_pending(self):
        if self.request_id == "":
            return False
        return self.status not in [
            ReportStatus.COMPLETED.value,
            ReportStatus.CANCELED.value,
        ]

    @property
    def is_active(self):
        if self.request_id == "":
            return False
        return self.status in [ReportStatus.ACTIVE.value]

    @property
    def iso_duration(self):
        """ Duration is sent in isodate duration format"""
        if not self.duration:
            # To accommodate the Kisensum VTN server, a null report duration has a special meaning
            # to the VEN: the report request should continue indefinitely, with no scheduled
            # completion time.
            # In this UpdateReport request, a null duration is sent as 0 seconds (one-time report)
            # to ensure that the UpdateReport has a valid construction.
            return "PT0S"
        return isodate.duration_isoformat(self.duration)


class TelemetryValues(db.Entity):
    """Model object for telemetry values."""

    created = orm.Required(datetime, default=datetime.utcnow)
    report_specifier_id = orm.Required(str)
    values = orm.Optional(orm.Json)
    start_time = orm.Optional(datetime, default=datetime.utcnow)
    end_time = orm.Optional(datetime)

    def __str__(self):
        """Format the instance as a string suitable for trace display."""
        tel_string = f"{self.__class__.__name__}: "
        tel_string += f"created_on:{self.created}; "
        tel_string += f"report_specifier_id:{self.report_specifier_id}; "
        tel_string += f"values:{self.values}; "
        tel_string += f"start_time:{self.start_time} "
        tel_string += f"end_time:{self.end_time} "
        return tel_string

    @property
    def values_dict(self) -> dict:
        return json.loads(self.values)

    @property
    def duration(self):
        if self.end_time == None:
            return None
        return self.end_time - self.start_time

    @property
    def iso_duration(self):
        """ Duration is sent in isodate duration format"""
        if not self.duration:
            return "PT0S"
        return isodate.duration_isoformat(self.duration)


def setup_db(filepath=None):
    if filepath:
        db.bind(provider="sqlite", filename=filepath, create_db=True)
    else:
        db.bind(provider="sqlite", filename=":memory:", create_db=True)
    db.generate_mapping(create_tables=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""The setup script."""
from setuptools import find_packages
from setuptools import setup

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "isodate",
    "lxml",
    "requests",
    "signxml",
    "pytz",
    "python-dateutil",
    "pony",
    "pem",
    "voluptuous",
    "bs4",
]

setup_requirements = ["pytest-runner"]

test_requirements = ["pytest", "factory-boy", "pytest-ponyorm", "pytest-cov"]

setup(
    author="Carbon Coop",
    author_email="peter@carbon.coop",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.7",
    ],
    description="OpenADR Virtual End Node",
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="pyoadr_ven",
    name="pyOADR-VEN",
    packages=find_packages(include=["pyoadr_ven"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://carboncoop.gitlab.io/pyoadr-ven/",
    version="0.4.8",
    zip_safe=False,
)
